import os

import numpy as np
import astropy.stats as astats

import click

import matplotlib as mpl
mpl.use('Agg')

import matplotlib.pyplot as plt
from matplotlib.backends import backend_pdf

from nenucal import datahandler
from nenucal.delayflag import make_ps
from ps_eor import psutil

mpl.rcParams['image.cmap'] = 'Spectral_r'
mpl.rcParams['image.origin'] = 'lower'
mpl.rcParams['image.interpolation'] = 'nearest'
mpl.rcParams['axes.grid'] = True


class GainSol(object):
    
    def __init__(self, time, freqs, ant, directions, pol, amp, phase):
        self.time = time
        self.freqs = freqs
        self.ant = ant
        self.dir = directions
        self.pol = pol
        self.amp = amp
        self.phase = phase
        self.d = self.amp * np.exp(1j * self.phase)
        

def open_sol(file_h5):
    import losoto.h5parm
    sol_file = losoto.h5parm.h5parm(file_h5)
    solset = sol_file.getSolsets()[0]
    soltab, soltab_phase = solset.getSoltabs(useCache=True)

    ant = soltab.getAxisValues('ant')
    directions = soltab.getAxisValues('dir')
    time = soltab.getAxisValues('time')
    pol = soltab.getAxisValues('pol')

    freqs = soltab.getAxisValues('freq')

    weight = soltab.getValues(weight=True)[0].astype(bool)
    amp = np.ma.array(soltab.getValues(weight=False)[0], mask=~weight)
    phase = np.ma.array(soltab_phase.getValues(weight=False)[0], mask=~weight)

    amp = amp[:, :, :]
    phase = phase[:, :, :]

    sol_file.close()

    return GainSol(time, freqs, ant, directions, pol, amp, phase)


# def get_mean_gain(gain_sols):
#     ants = sorted(set([a for gain_sol in gain_sols for a in gain_sol.ant]))
#     gain_sols_m = np.zeros((1, len(gain_sols[0].freqs), len(ants), 1, 2), dtype=np.complex)
#     for i, ant in enumerate(ants):
#         d = np.concatenate([gain_sol.d[:, :, gain_sol.ant == ant, 0, :] for gain_sol in gain_sols if ant in gain_sol.ant])
#         gain_sols_m[0, :, i] = np.mean(astats.sigma_clip(d, axis=0), axis=0)

#     return gain_sols_m


def get_mean_gain(gain_sols):
    ants = sorted(set([a for gain_sol in gain_sols for a in gain_sol.ant]))
    gain_sols_m = np.zeros((1, len(gain_sols[0].freqs), len(ants), 1, 2), dtype=complex)
    for i, ant in enumerate(ants):
        ant_gain_sols = []
        for gain_sol in gain_sols:
            if ant in gain_sol.ant:
                d = gain_sol.d[:, :, gain_sol.ant == ant, 0]
                d = astats.sigma_clip(d.real, maxiters=10, sigma=5) + 1j * astats.sigma_clip(d.imag, maxiters=10, sigma=5)
                time_mask = astats.sigma_clip(d.std(axis=(1, 2, 3)), sigma=2).mask
                time_mask2 = astats.sigma_clip(d.mask.sum(axis=(1, 2, 3)), sigma=2).mask
                ant_gain_sols.append(d[~(time_mask | time_mask2), :, 0])

        gain_sols_m[0, :, i, 0] = np.mean(np.concatenate(ant_gain_sols), axis=0)

    return gain_sols_m


def save_h5(h5_file, time, freq, amps, phases, ant, direction='Main', pol=[]):
    import losoto.h5parm
    weights = np.ones_like(amps)
    
    print(ant)

    axis_vals = [time, freq, ant, ['[%s]' % direction], np.array(['XX', 'YY'])]
    axes_name = ['time', 'freq', 'ant', 'dir', 'pol']

    if os.path.exists(h5_file):
        os.remove(h5_file)
        
    print(time.shape, freq.shape, ant.shape, amps.shape, phases.shape)

    los_h5 = losoto.h5parm.h5parm(h5_file, readonly=False)
    try:
        los_h5.makeSolset('sol000')
        sol_set = los_h5.getSolset('sol000')
        sol_set.makeSoltab('amplitude', 'amplitude000', axesNames=axes_name, axesVals=axis_vals,
                           vals=amps, weights=weights)
        sol_set.makeSoltab('phase', 'phase000', axesNames=axes_name, axesVals=axis_vals, vals=phases, 
                           weights=weights)
    finally:
        los_h5.close()


def plot(freqs, gain_sols_m, ants, output_pdf, nrows=5):
    pdf = backend_pdf.PdfPages(output_pdf)

    for i in np.arange(np.ceil(gain_sols_m.shape[2] / nrows)):
        fig, axs = plt.subplots(figsize=(10, 15), nrows=nrows, sharex=True)
        for j, ax in enumerate(axs):
            k = int(i * nrows + j)
            if k >= gain_sols_m.shape[2]:
                continue
            ax.plot(freqs * 1e-6, np.abs(gain_sols_m[0, :, k, 0, :]))
            ax.text(0.05, 0.95, ants[k], transform=ax.transAxes, va='top', ha='left')
        axs[-1].set_xlabel('Frequency [MHz]')
        fig.tight_layout(pad=0)
        pdf.savefig(fig)
    pdf.close()


def plot_delay(freqs, gain_sols_m, ants, output_pdf, fmax=71e6, ncols=5, delay_max=7):
    nrows = int(np.ceil(gain_sols_m.shape[2] / 5))
    fig, axs = plt.subplots(figsize=(10, nrows * 2.25), ncols=ncols, nrows=nrows, sharex=True, sharey=True)
    axs = axs.flatten()
    idx = (freqs < fmax)
    gain_sols_m[np.isnan(gain_sols_m)] = 0
    for i, (ant, ax) in enumerate(zip(ants, axs)):
        delay = psutil.get_delay(freqs[idx]) * 1e6
        ps = make_ps(gain_sols_m[0, idx, i, 0, :], freqs[idx])
        ps = ps / ps.max()
        ax.plot(delay, ps)
        ax.set_xlim(0, delay_max)
        ax.set_yscale('log')
        ax.text(0.05, 0.95, ant, transform=ax.transAxes, va='top', ha='left')
    fig.tight_layout(pad=0)
    fig.savefig(output_pdf)


@click.command()
@click.argument('obs_ids', type=str)
@click.argument('output_h5', type=str)
@click.option('--config', '-c', help='Data handler configuration file', type=str,
              default='data_handler.toml', show_default=True)
@click.option('--input_h5_name', '-i', help='Name of the input h5 file', type=str,
              default='instrument_di_bp.h5', show_default=True)
@click.option('--data_level', '-l', help='Level of the data', type=str, default='L2_12C40S', show_default=True)
def main(obs_ids, output_h5, config, input_h5_name, data_level):
    dh = datahandler.DataHandler.from_file(config)

    all_gain_sols_m = []
    all_freqs = []
    obs_ids = dh.get_obs_ids(obs_ids)

    for sw in dh.get_spectral_windows():
        h5_files = [f'{k}/{input_h5_name}' for k in dh.get_all_ms_path(obs_ids, [data_level], [sw])]
        gain_sols = [open_sol(h5_file) for h5_file in h5_files]
        sw_gain_sols_m = get_mean_gain(gain_sols)
        all_gain_sols_m.append(sw_gain_sols_m)
        all_freqs.append(gain_sols[0].freqs)

    gain_sols_m = np.concatenate(all_gain_sols_m, axis=1)
    freqs = np.concatenate(all_freqs)
    ants = sorted(set([a for gain_sol in gain_sols for a in gain_sol.ant]))

    plot(freqs, gain_sols_m, ants, output_h5.replace('.h5', '') + '.pdf')
    plot_delay(freqs, gain_sols_m, ants, output_h5.replace('.h5', '') + '_delay.pdf')

    save_h5(output_h5, gain_sols[0].time[:1], freqs, np.abs(gain_sols_m), 
            np.angle(gain_sols_m), np.array(ants))


if __name__ == '__main__':
    main()



#!/usr/bin/env python

import os
import sys
import fnmatch

import toml
import click

from nenucal import tasks
from nenucal import datahandler
from nenucal import __version__
from nenucal.settings import Settings

t_file = click.Path(exists=True, dir_okay=False)
t_dir = click.Path(exists=True, file_okay=False)

tasks_help = '\n'.join([f'  {name.ljust(17)}{desc}' for name, desc in tasks.get_all_tasks_descriptions().items()])


@click.command()
@click.version_option(__version__)
@click.argument('config_file', type=click.Path(exists=True, dir_okay=False))
@click.argument('ms_ins_or_obs_ids', nargs=-1, required=True)
def main(config_file, ms_ins_or_obs_ids):
    ''' Calibration pipeline

        \b
        CONFIG_FILE: Configuration file
        MS_INS_OR_OBS_IDS: Measurement sets to process or OBS_IDS in case you have set data_handler in CONFIG_FILE
    '''
    try:
        s = Settings.load_with_defaults(config_file, with_type_keyword=True, check_args=False)
    except toml.TomlDecodeError as e:
        print(f'Error parsing configuration: {e}')
        sys.exit(1)

    s_default = Settings.get_defaults()

    if 'steps' not in s:
        print(f'Error: a "steps" list is required in {config_file}')
        sys.exit(1)

    if 'worker' not in s:
        s['worker'] = s_default['worker']
        # print(f'Error: a "worker" section is required in {config_file}')
        # sys.exit(1)

    if 'sky_model' not in s:
        s['sky_model'] = s_default['sky_model']
        # print(f'Error: a "sky_model" section is required in {config_file}')
        # sys.exit(1)

    if 'data_handler' in s and "config_file" in s["data_handler"] and s["data_handler"]["config_file"]:
        print(f'Using {s["data_handler"]["config_file"]} data handler configuration')
        dh = datahandler.DataHandler.from_file(s['data_handler']['config_file'])
        level = s['data_handler']['data_level']
        ms_ins = []
        for obs_id_str in ms_ins_or_obs_ids:
            obs_ids, sws = dh.get_obs_ids_and_spectral_windows(obs_id_str)
            ms_ins.extend(dh.get_all_ms_path(obs_ids, [level], sws))
    else:
        ms_ins = ms_ins_or_obs_ids

    all_tasks = tasks.get_all_tasks()

    for step in s.steps:
        if step not in s.keys():
            print(f'Error: step {step} is not known in {config_file}')
            sys.exit(1)
        if not ('type' in s[step].keys() and s[step].type in all_tasks) and step not in all_tasks:
            print(f'Error: type not given or unknown for step {step} in {config_file}')
            print(f'Possible type: {",".join(all_tasks)}')
            sys.exit(1)

    for step in s.steps:
        if 'type' in s[step].keys() and s[step].type in all_tasks:
            t = s[step].type
        elif step in all_tasks:
            t = step

        ms_ins = all_tasks[t](s[step], s.worker, s.sky_model).run(ms_ins)

        if not ms_ins:
            print('\nError: no MS to process.')
            sys.exit(1)



#!/usr/bin/env python

import os
import sys
import itertools
from datetime import datetime
from collections import defaultdict

import toml
import click
from tabulate import tabulate
from click import confirm, style

import numpy as np

from nenucal import __version__, msutils
from nenucal import datahandler

from libpipe import worker, futils

t_file = click.Path(exists=True, dir_okay=False)

ms_file_to_check = 'table.info'


@click.group()
@click.version_option(__version__)
def main():
    ''' NenuFAR-CD data management utilities ...'''


@main.command('list')
@click.argument('obs_ids', type=str)
@click.option('--config', '-c', help='Data handler configuration file', type=t_file, default='data_handler.toml')
def list(obs_ids, config):
    ''' List all obs_ids '''
    dh = datahandler.DataHandler.from_file(config)

    header = ['Obs_id'] + [*dh.get_levels()]
    data = []

    for obs_id in dh.get_obs_ids(obs_ids):
        o = [obs_id]
        for level in dh.get_levels():
            counts = []
            for sw in dh.get_spectral_windows():
                mss = dh.get_ms_path(obs_id, level, sw)
                n_mss = len(mss)
                n_mss_exists = sum([os.path.exists(ms + f'/{ms_file_to_check}') for ms in mss])
                if n_mss == n_mss_exists:
                    counts.append(style(f'{n_mss}', fg='green'))
                elif n_mss_exists > 0:
                    counts.append(style(f'{n_mss_exists}', fg='yellow'))
                else:
                    counts.append(style(f'0', fg='red'))
            o.append(','.join(counts))
        data.append(o)

    print(tabulate(data, header))


@main.command('add')
@click.argument('obs_id', type=str)
@click.argument('nodes_distribution_id', type=str)
@click.option('--config', '-c', help='Data handler configuration file', type=t_file, default='data_handler.toml')
def add(obs_id, nodes_distribution_id, config):
    ''' Add obs_id '''
    dh = datahandler.DataHandler.from_file(config)

    if nodes_distribution_id not in dh.get_nodes_distribution():
        print(f'Error: nodes distribution ID {nodes_distribution_id} does not exist.')
        sys.exit(0)

    if obs_id in dh.get_obs_ids():
        print(f'Error: observation ID {obs_id} already exist.')
        sys.exit(0)

    nodes = dh.get_nodes_distribution()[nodes_distribution_id]

    if confirm(f'Adding {obs_id} with nodes {",".join(nodes)}'):
        s = toml.load(config, _dict=defaultdict)
        s['obs_ids'][obs_id] = nodes
        f = f'{config}.{datetime.now().strftime("%Y%m%d")}'
        backup_config = f
        i = 1
        while os.path.exists(backup_config):
            backup_config = f'{f}.{i}'
            i += 1

        os.rename(config, backup_config)

        with open(config, mode='w') as f:
            toml.dump(dict(s), f)

        print('Done !')

@main.command('add_all')
@click.argument('obs_ids', type=str, nargs=-1)
@click.option('--config', '-c', help='Data handler configuration file', type=t_file, default='data_handler.toml')
@click.option('--skip_existing', '-s', help='Skip existing OBS_ID', is_flag=True)
def add_all(obs_ids, config, skip_existing):
    ''' Add all obs_id '''
    dh = datahandler.DataHandler.from_file(config)
    s = toml.load(config, _dict=defaultdict)
    all_obs_ids = dh.get_obs_ids()

    for obs_id, nodes_distribution_id in zip(obs_ids, itertools.cycle(dh.get_nodes_distribution().keys())):
        if obs_id in all_obs_ids:
            print(f'Error: observation ID {obs_id} already exist.')
            if not skip_existing:
                sys.exit(0)

        nodes = dh.get_nodes_distribution()[nodes_distribution_id]
        print(f'Adding {obs_id} with nodes {",".join(nodes)} (ID: {nodes_distribution_id})')
        s['obs_ids'][obs_id] = nodes

    if len(obs_ids) > 0:
        if confirm('Confirm update ?'):
            f = f'{config}.{datetime.now().strftime("%Y%m%d")}'
            backup_config = f
            i = 1
            while os.path.exists(backup_config):
                backup_config = f'{f}.{i}'
                i += 1

            os.rename(config, backup_config)

            with open(config, mode='w') as f:
                toml.dump(dict(s), f)

            print('Done !')
        else:
            print('Changes discarded.')


@main.command('get_obs_ids')
@click.argument('obs_ids', type=str)
@click.option('--config', '-c', help='Data handler configuration file', type=t_file, default='data_handler.toml')
def get_obs_ids(obs_ids, config):
    ''' List all obs_ids (without existence check) '''
    dh = datahandler.DataHandler.from_file(config)

    print('\n'.join(dh.get_obs_ids(obs_ids)))


@main.command('get_ms')
@click.argument('level', type=str)
@click.argument('obs_ids', type=str)
@click.option('--config', '-c', help='Data handler configuration file', type=t_file, default='data_handler.toml')
@click.option('--sws', '-s', help='Spectral windows', type=str, default='all')
def get_ms(level, obs_ids, config, sws):
    ''' Return a list of all MS corresponding to given OBS_IDS and SWS '''
    dh = datahandler.DataHandler.from_file(config)

    if sws == 'all':
        sws = dh.get_spectral_windows()
    else:
        sws = [k.upper() for k in sws.split(',')]

    for obs_id in dh.get_obs_ids(obs_ids):
        for sw in sws:
            print(' '.join(dh.get_ms_path(obs_id, level, sw)), end=' ')


@main.command('remove')
@click.argument('level', type=str)
@click.argument('obs_ids', type=str)
@click.option('--config', '-c', help='Data handler configuration file', type=t_file, default='data_handler.toml')
@click.option('--sws', '-s', help='Spectral windows', type=str, default='all')
def remove(level, obs_ids, config, sws):
    ''' Remove all MS corresponding to given OBS_IDS and SWS '''
    dh = datahandler.DataHandler.from_file(config)

    if sws == 'all':
        sws = dh.get_spectral_windows()
    else:
        sws = [k.upper() for k in sws.split(',')]

    obs_ids = dh.get_obs_ids(obs_ids)

    if confirm(style(f'Removing obs_ids {",".join(obs_ids)} for SW {",".join(sws)} ?', fg='yellow')):
        for obs_id in obs_ids:
            for sw in sws:
                for ms in dh.get_ms_path(obs_id, level, sw):
                    futils.rm_if_exist(ms, verbose=True)
    else:
        print('No changes made.')


@main.command('retrieve')
@click.argument('remote_host', type=str)
@click.argument('obs_ids', type=str)
@click.option('--config', '-c', help='Data handler configuration file', type=t_file, default='data_handler.toml')
@click.option('--dry_run', help='Run in dry mode', is_flag=True)
def retrieve(obs_ids, remote_host, config, dry_run):
    ''' Return a list of all MS corresponding to given OBS_IDS and SWS '''
    dh = datahandler.DataHandler.from_file(config)
    assert remote_host in dh.get_remote_hosts(), f'Remote host {remote_host} needs to be defined in {config}'

    target_level = dh.get_remote_level(remote_host)
    target_host = dh.get_remote_host(remote_host)
    target_password_file = dh.get_remote_password_file(remote_host)

    assert target_level in dh.get_levels(), f'{target_level} data level can be retrieved'

    w = worker.WorkerPool(dh.get_all_hosts(), name='Transfert', max_tasks_per_worker=1, debug=dry_run, dry_run=dry_run)

    for obs_id in dh.get_obs_ids(obs_ids, include_n2_obs_ids=False):
        date = obs_id.split('_')[0]
        year = date[:4]
        month = date[4:6]

        remote_path = dh.get_remote_data_path(remote_host).replace('%YEAR%', year).replace('%MONTH%', month)
        remote_path = remote_path.replace('%OBS_ID%', obs_id)

        for sw in dh.get_spectral_windows():
            files = ' '.join([f'{target_host}:{remote_path}/SB{sb}.MS' for sb in dh.get_sbs(sw)])
            target = dh.get_dir_path(obs_id, target_level, sw)
            node = dh.get_node(obs_id, sw)

            for i in range(100):
                log_file = f'{target}/transfert_{i}.log'
                if not os.path.exists(log_file):
                    break

            if not os.path.exists(target):
                os.makedirs(target)

            cmd = f'sshpass -f {target_password_file} rsync --progress -v -am {files} {target}'

            w.add(cmd, run_on_host=node, output_file=log_file)

    w.execute()


@main.command('l1_to_l2')
@click.argument('level', type=str)
@click.argument('obs_ids', type=str)
@click.option('--config', '-c', help='Data handler configuration file', type=t_file, default='data_handler.toml')
@click.option('--force', help='Force overwrite data if already exists', is_flag=True)
@click.option('--l1_level', help='L1 level name', type=str, default='L1')
@click.option('--max_concurrent', '-m', help='Maximum concurrent tasks on a node', type=int, default=1)
@click.option('--dry_run', help='Do not do anything', is_flag=True)
def l1_to_l2(level, obs_ids, config, force, l1_level, max_concurrent, dry_run):
    ''' Create L2 data (at level LEVEL) from L1 data for the given OBS_IDS'''
    dh = datahandler.DataHandler.from_file(config)
    assert 'L1' in dh.get_levels(), f'L1 data level needs to be defined'
    assert level in dh.get_levels(), f'{level} data level needs to be defined'

    dppp_file = dh.get_l1_to_l2_config(level)

    w = worker.WorkerPool(dh.get_all_hosts(), name='L1 to L2', max_tasks_per_worker=max_concurrent,
                          debug=dry_run, dry_run=dry_run)

    obs_ids, sws = dh.get_obs_ids_and_spectral_windows(obs_ids)

    for obs_id in obs_ids:
        for sw in sws:
            msins_list = dh.get_ms_path(obs_id, l1_level, sw)
            msins = ','.join(msins_list)
            spec_ms_process = []

            if dh.is_n2_obs_id(obs_id):
                ms_info = msutils.get_info_from_ms_files(msins_list)
                nodes = dh.get_n2_nodes(obs_id)
                msouts = dh.get_ms_path(obs_id, level, sw)
                n = len(nodes)
                if l1_level.upper().startswith('L2'):
                    spec_ms_process = zip(nodes, msins_list, msouts, [0] * n, [0] * n)
                else:
                    n_slots = (ms_info['end_time'] - ms_info['start_time']) // ms_info['int_time']
                    starttimeslots = np.linspace(0, n_slots, n + 1).astype(int)
                    n_times_slices = np.diff(starttimeslots)
                    starttimeslots = starttimeslots[:-1]
                    spec_ms_process = zip(nodes, [msins] * n, msouts, starttimeslots, n_times_slices)
            else:
                msout = dh.get_ms_path(obs_id, level, sw)[0]
                node = dh.get_node(obs_id, sw)
                spec_ms_process.append((node, msins, msout, 0, 0))

            for node, msins, msout, starttimeslot, ntimes in spec_ms_process:
                target, msout_name = os.path.split(msout)

                if os.path.exists(msout):
                    if force:
                        print(f'Warning: {msout} already exists')
                    else:
                        print(f'Error: {msout} already exists')
                        return 1

                if not os.path.exists(target):
                    os.makedirs(target)

                for i in range(1000):
                    log_file = f'{target}/l2_to_l1_{msout_name}_{i}.log'
                    if not os.path.exists(log_file):
                        break

                cmd = f'DPPP {os.path.abspath(dppp_file)} msin=[{msins}] msout={msout} msout.overwrite=true'
                if starttimeslot!= 0 or ntimes != 0:
                    cmd += f' msin.starttimeslot={starttimeslot} msin.ntimes={ntimes}'

                w.add(cmd, run_on_host=node, output_file=log_file)

    w.execute()


@main.command('make_ms_list')
@click.argument('level', type=str)
@click.argument('obs_ids', type=str)
@click.option('--config', '-c', help='Data handler configuration file', type=t_file, default='data_handler.toml')
@click.option('--target', help='Target directory', default='ms_lists')
def make_ms_list(level, obs_ids, config, target):
    ''' Make MS list for all OBS_IDS at level LEVEL '''
    dh = datahandler.DataHandler.from_file(config)
    assert level in dh.get_levels(), f'{level} data level needs to be defined'

    if not os.path.exists(target):
        os.makedirs(target)

    for obs_id in dh.get_obs_ids(obs_ids):
        for sw in dh.get_spectral_windows():
            ms = dh.get_ms_path(obs_id, level, sw)[0]
            filepath = os.path.join(target, f'{obs_id}_{level}_{sw}')

            with open(filepath, 'w') as f:
                f.write(f'{ms}\n')



import os
import re
import pathlib
import datetime

import numpy as np

import lsmtool
lsmtool.logger.setLevel('warning')

from libpipe import worker, futils

from . import utils, flagutils, msutils
from .settings import TEMPLATE_DIR

cwd = os.getcwd()


def get_cal_config(name):
    return pathlib.Path(__file__).absolute().parent / 'cal_config' / name


class SkyModel(object):

    def __init__(self, sky_model_name='app_sky_model'):
        self.sky_model_name = sky_model_name
        self.stat_cache = dict()

    def get_sky_model(self, msin):
        return f'{msin}/sky_model/{self.sky_model_name}.sourcedb'

    def get_sky_model_bbs(self, msin):
        return f'{msin}/sky_model/{self.sky_model_name}.skymodel'

    def get_stat(self, msin):
        if msin not in self.stat_cache:
            model = lsmtool.load(self.get_sky_model_bbs(msin))
            s = dict(zip(model.getPatchNames(), model.getColValues('I', aggregate='sum'))), model.getPatchPositions()
            self.stat_cache[msin] = s
        return self.stat_cache[msin]

    def get_patch_i(self, msin, patch):
        if isinstance(patch, str):
            return self.get_stat(msin)[0][patch]
        return sum(self.get_patch_i(msin, k) for k in patch)

    def get_patch_coord(self, msin, patch):
        return self.get_stat(msin)[1][patch]

    def get_patchs(self, msin, include_main=True, exclude=None):
        ateams = self.get_stat(msin)[0]
        patches = sorted(ateams, key=ateams.get, reverse=True)
        if 'Main' in patches:
            patches.remove('Main')
            if include_main:
                patches.append('Main')
        if exclude:
            for patch in exclude:
                if patch in patches:
                    patches.remove(patch)
        return patches

    def get_directions(self, msin, dir_name_or_idx):
        patchs = self.get_patchs(msin)
        dirs = []
        if isinstance(dir_name_or_idx, (list, np.ndarray, tuple)):
            dirs = list(set(patchs).intersection(set(dir_name_or_idx)))
            if len(dirs) != len(dir_name_or_idx):
                print(f'Warning: not all directions were recognized: {dir_name_or_idx} != {dirs}')
        elif isinstance(dir_name_or_idx, (int, np.int64, slice)):
            dirs = patchs[dir_name_or_idx]
        elif dir_name_or_idx == 'all':
            dirs = patchs
        elif isinstance(dir_name_or_idx, str) and dir_name_or_idx in patchs:
            dirs = [dir_name_or_idx]
        elif isinstance(dir_name_or_idx, str) and dir_name_or_idx.startswith('!'):
            if dir_name_or_idx[1:] in patchs:
                dirs = patchs.copy()
                dirs.remove(dir_name_or_idx[1:])
        return dirs

    def copy(self, msins, msouts):
        futils.zip_copytree(msins, msouts, self.get_sky_model(''))
        futils.zip_copy(msins, msouts, self.get_sky_model_bbs(''))


class CalSettings(object):

    def __init__(self, parmdb='instrument.h5', sol_int=1, sol_int_flux_per_slot_per_sec=0,
                 sol_int_min=2, sol_int_max=120, mode='fulljones', uv_min=None, smoothnessconstraint=4e6,
                 extra={}, **kargs):
        self.parmdb = parmdb
        self.sol_int = sol_int
        self.cal_mode = mode
        self.uv_min = uv_min
        self.extra_params = extra
        self.sol_int_min = sol_int_min
        self.sol_int_max = sol_int_max
        self.sol_int_flux_per_slot_per_sec = sol_int_flux_per_slot_per_sec
        self.smoothnessconstraint = smoothnessconstraint
        self.freq_cache = {}
        self.interval_cache = {}

    def get_time_interval(self, msin):
        if msin not in self.interval_cache:
            self.interval_cache[msin] = msutils.get_ms_time_interval(msin)
        return self.interval_cache[msin]

    def get_freq(self, msin):
        if msin not in self.freq_cache:
            self.freq_cache[msin] = msutils.get_ms_freqs(msin)[0].mean()
        return self.freq_cache[msin]

    def get_uv_min(self, msin):
        if not self.uv_min:
            return self.uv_min
        if isinstance(self.uv_min, (float, int)):
            return self.uv_min
        fmhz = self.get_freq(msin) * 1e-6
        fmhzs, lims = list(zip(*sorted(self.uv_min.items())))
        fmhzs = [int(k) for k in fmhzs]
        i = np.where(np.array(fmhzs) - fmhz >= 0)[0][0]
        return lims[i]

    def get_sol_int(self, msin, patch_i):
        if self.sol_int_flux_per_slot_per_sec is 0:
            return int(self.sol_int)

        int_time = self.get_time_interval(msin)
        b = (self.sol_int_flux_per_slot_per_sec / patch_i / int_time)
        if b >= self.sol_int_max:
            return self.sol_int_max
        f = utils.factors(self.sol_int_max)
        c = f[(f - b) > 0][0]
        return int(max(self.sol_int_min, c))


class MultiCommands(object):

    def __init__(self, worker_settings, name, exec_name, max_time=None):
        self.worker_settings = worker_settings
        self.exec_name = exec_name
        self.name = name
        self.max_time = max_time
        if self.max_time == 0:
            self.max_time = None

    def build_command(self, msin):
        parameters = self.get_parameters(msin)
        if parameters is None:
            return None
        cmd = f'cd {cwd}; {self.exec_name} {" ".join(parameters)}'
        return cmd

    def get_parameters(self, msin):
        return []

    def get_out_file(self, msin):
        return msin

    def get_log_file(self, msin):
        return None

    def run(self, in_files):
        print(f'Starting {self.name} ...')
        kargs = self.worker_settings.copy()
        del kargs['run_on_file_host']
        del kargs['run_on_file_host_pattern']
        pool = worker.get_worker_pool(self.name, max_time=self.max_time, **kargs)
        out_files = []

        for in_file in in_files:
            host = None
            if self.worker_settings.run_on_file_host and self.worker_settings.run_on_file_host_pattern:
                r = re.search(self.worker_settings.run_on_file_host_pattern, in_file)
                if r is not None:
                    host = r.group(1)
            in_file = os.path.normpath(in_file)
            out_file = self.get_out_file(in_file)
            cmd = self.build_command(in_file)
            if cmd is None:
                print(f'Skipping {in_file}')
                continue
            pool.add(cmd, output_file=self.get_log_file(in_file), run_on_host=host)
            out_files.append(out_file)

        pool.execute()

        return out_files


class AoQuality(MultiCommands):

    def __init__(self, worker_settings, corrected_data=True, max_time=None):
        MultiCommands.__init__(self, worker_settings, 'AoQuality', 'aoquality collect', max_time=max_time)
        self.corrected_data = corrected_data

    def get_parameters(self, msin):
        p = [msin]
        if self.corrected_data:
            p.insert(0, '-d CORRECTED_DATA')
        return p


class MakeAppSkyModel(MultiCommands):

    def __init__(self, worker_settings, app_sky_model_name, int_sky_models,
                 build_sky_model_settings):
        MultiCommands.__init__(self, worker_settings, 'MakeAppSkyModel', 'modeltool attenuate')
        self.app_sky_model_name = app_sky_model_name
        self.int_sky_models = int_sky_models
        self.s_bsm = build_sky_model_settings

    def get_parameters(self, msin):
        out_file = SkyModel(self.app_sky_model_name).get_sky_model_bbs(msin)
        futils.mkdir(os.path.dirname(out_file))

        p = [msin, " ".join(self.int_sky_models),
             f'-m {self.s_bsm.min_flux}',
             f'-p {self.s_bsm.min_flux_path}',
             f'-e {self.s_bsm.ateam_min_elevation}',
             f'-o {out_file}']

        for patch in self.s_bsm.ateam_always_keep:
            p.append(f'-k {patch}')

        for patch in self.s_bsm.ateam_remove:
            p.append(f'-r {patch}')

        return p


class ConcatenateSkyModel(MultiCommands):

    def __init__(self, worker_settings, output_sky_model_name, input_sky_model_names,
                 build_sky_model_settings):
        MultiCommands.__init__(self, worker_settings, 'ConcatenateSkyModel', 'modeltool concatenate')
        self.output_sky_model_name = output_sky_model_name
        self.input_sky_model_names = input_sky_model_names
        self.s_bsm = build_sky_model_settings

    def get_parameters(self, msin):
        out_file = SkyModel(self.output_sky_model_name).get_sky_model_bbs(msin)
        inp_files = [SkyModel(name).get_sky_model_bbs(msin) for name in self.input_sky_model_names]
        futils.mkdir(os.path.dirname(out_file))

        p = [msin, " ".join(inp_files),
             f'-m {self.s_bsm.min_flux}',
             f'-p {self.s_bsm.min_flux_path}',
             f'-e {self.s_bsm.ateam_min_elevation}',
             f'-o {out_file}']

        for patch in self.s_bsm.ateam_always_keep:
            p.append(f'-k {patch}')

        for patch in self.s_bsm.ateam_remove:
            p.append(f'-r {patch}')

        return p


class MakeSourceDB(MultiCommands):

    def __init__(self, worker_settings, app_sky_model_name):
        MultiCommands.__init__(self, worker_settings, 'MakeSourceDB', 'makesourcedb')
        self.app_sky_model_name = app_sky_model_name

    def get_parameters(self, msin):
        sky_model = SkyModel(self.app_sky_model_name)
        return [f'in={sky_model.get_sky_model_bbs(msin)}', f'out={sky_model.get_sky_model(msin)}', 'append=false']


class PlotSolutions(MultiCommands):

    def __init__(self, worker_settings, parmdb_in_name, clip=False):
        self.parmdb_in_name = parmdb_in_name
        self.clip = clip
        MultiCommands.__init__(self, worker_settings, 'PlotSol', 'soltool plot')

    def get_parameters(self, msin):
        parmdb = f'{msin}/{self.parmdb_in_name}'
        if not os.path.exists(parmdb):
            return None
        p = [parmdb, f'--plot_dir=plots_{os.path.basename(self.parmdb_in_name.split(".")[0])}']
        if self.clip:
            p.append('--clip')
        return p


class SmoothSolutions(MultiCommands):

    def __init__(self, worker_settings, parmdb_in_name, time_min=15, freq_mhz=2, main_time_min=None,
                 main_freq_mhz=None, clip_nsigma=4):
        self.parmdb_in_name = parmdb_in_name
        self.fwhm_time = time_min
        self.fwhm_freq = freq_mhz
        self.main_fwhm_time = main_time_min
        self.main_fwhm_freq = main_freq_mhz

        if self.main_fwhm_freq is None:
            self.main_fwhm_freq = freq_mhz
        if self.main_fwhm_time is None:
            self.main_fwhm_time = freq_mhz

        self.clip_nsigma = clip_nsigma
        MultiCommands.__init__(self, worker_settings, 'SmoothSol', 'soltool smooth')

    def get_parameters(self, msin):
        parmdb = f'{msin}/{self.parmdb_in_name}'
        if not os.path.exists(parmdb):
            return None

        p = [parmdb, f'--fwhm_time={self.fwhm_time}', f'--fwhm_freq={self.fwhm_freq}',
             f'--main_fwhm_time={self.main_fwhm_time}', f'--main_fwhm_freq={self.main_fwhm_freq}',
             f'--clip_nsigma={self.clip_nsigma}']
        return p


class CopyFlag(MultiCommands):

    def __init__(self, worker_settings, in_postfix, out_postfix):
        self.in_postfix = in_postfix
        self.out_postfix = out_postfix
        MultiCommands.__init__(self, worker_settings, 'CopyFlag', 'flagtool copy')

    def get_parameters(self, msin):
        ms_flag = f'_{self.in_postfix}.MS'.join(msin.rsplit(f'_{self.out_postfix}.MS', 1))
        return [ms_flag, msin]


class RestoreOrBackupFlag(MultiCommands):

    def __init__(self, worker_settings, flag_name='pre_cal_flags.h5'):
        self.flag_name = flag_name
        MultiCommands.__init__(self, worker_settings, 'RestoreOrBackupFlag', 'flagtool')

    def get_parameters(self, msin):
        flag_file = os.path.join(msin, self.flag_name)
        if os.path.exists(flag_file):
            return ['restore', msin, flag_file]
        return ['backup', msin, flag_file]


class DPPP(MultiCommands):

    def __init__(self, worker_settings, name, parset, max_time=None):
        self.parset = parset
        MultiCommands.__init__(self, worker_settings, name, 'DPPP', max_time=max_time)

    def get_log_file(self, msin):
        log_dir = os.path.join(os.path.dirname(msin), 'logs')
        if not os.path.exists(log_dir):
            os.makedirs(log_dir)
        log_name = self.name.replace(' ', '_').lower()
        date_str = datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%S')
        return f'{log_dir}/{date_str}_{os.path.basename(msin)}_{log_name}.log'

    def build_command(self, msin):
        parameters = self.get_parameters(msin)
        if parameters is None:
            return None
        return f'cd {cwd}; {self.exec_name} {self.parset} msin={msin} {" ".join(parameters)}'

    def get_directions_str(self, directions):
        if isinstance(directions, str):
            return f'[[{directions}]]'
        return '[%s]' % ','.join(['[%s]' % k for k in directions])

    def get_directions_str_single_list(self, directions):
        if isinstance(directions, str):
            return f'[{directions}]'
        return '[%s]' % ','.join(['%s' % k for k in directions])


class CopyDataCol(DPPP):

    def __init__(self, worker_settings, col_in, col_out):
        self.col_in = col_in
        self.col_out = col_out
        DPPP.__init__(self, worker_settings, 'CopyCol', '')

    def get_parameters(self, msin):
        return ['numthreads=200', f'msin.datacolumn={self.col_in}',
                f'msout=.', f'msout.datacolumn={self.col_out}', 'steps=[]']


class CopyMS(DPPP):

    def __init__(self, worker_settings, col_in, ms_out_postfix):
        self.col_in = col_in
        self.postfix = ms_out_postfix
        DPPP.__init__(self, worker_settings, 'CopyMS', '')

    def get_out_file(self, msin):
        return f'_{self.postfix}.MS'.join(msin.rsplit('.MS', 1))

    def get_parameters(self, msin):
        msout = self.get_out_file(msin)
        return ['numthreads=200', 'msout.overwrite=true', f'msin.datacolumn={self.col_in}',
                f'msout={msout}', f'msout.datacolumn=DATA', 'steps=[]']


class DDEcal(DPPP):

    def __init__(self, worker_settings, cal_settings, sky_model, data_col='DATA', directions='all'):
        self.settings = cal_settings
        self.sky_model = sky_model
        self.directions = directions
        self.data_col = data_col
        self.data_col_out = data_col

        DPPP.__init__(self, worker_settings, 'DDEcal', get_cal_config('dppp_ddecal.parset'))

    def get_parameters(self, msin):
        parmdb_out = f'{msin}/{self.settings.parmdb}'

        if os.path.exists(parmdb_out):
            os.remove(parmdb_out)

        directions = self.sky_model.get_directions(msin, self.directions)

        sol_int = self.settings.get_sol_int(msin, self.sky_model.get_patch_i(msin, directions))

        p = []
        p.append(f'msin.datacolumn={self.data_col}')
        p.append(f'msout.datacolumn={self.data_col_out}')
        p.append(f'cal.sourcedb={self.sky_model.get_sky_model(msin)}')
        p.append(f'cal.directions={self.get_directions_str(directions)}')
        p.append(f'cal.h5parm={parmdb_out}')
        p.append(f'cal.solint={sol_int}')
        p.append(f'cal.mode={self.settings.cal_mode}')
        p.append(f'cal.smoothnessconstraint={self.settings.smoothnessconstraint}')

        for k, v in self.settings.extra_params.items():
            p.append(f'cal.{k}={v}')

        if self.settings.uv_min:
            p.append(f'cal.uvlambdamin={self.settings.get_uv_min(msin)}')

        return p


class DDEcalAvg(DDEcal):

    def __init__(self, worker_settings, cal_settings, sky_model, time_avg=4,
                 freq_avg=1, data_col='DATA', directions='Main'):
        DDEcal.__init__(self, worker_settings, cal_settings, sky_model, data_col=data_col, directions=directions)
        self.data_col_out = 'DATA'
        self.time_avg = time_avg
        self.freq_avg = freq_avg

    def get_out_file(self, msin):
        p = pathlib.Path(msin)
        return str(p.parent / ('tmp_' + p.name))

    def get_parameters(self, msin):
        msout = self.get_out_file(msin)

        p = DDEcal.get_parameters(self, msin)
        p.append('steps=[avg,cal]')
        p.append('avg.type=averager')
        p.append(f'avg.timestep={self.time_avg}')
        p.append(f'avg.freqstep={self.freq_avg}')
        p.append(f'msout={msout}')
        p.append(f'msout.overwrite=true')

        return p


class Predict(DPPP):

    def __init__(self, worker_settings, cal_settings, sky_model, col_out, directions='all', max_time=None):
        self.settings = cal_settings
        self.sky_model = sky_model
        self.col_out = col_out
        self.directions = directions

        if not self.settings.parmdb.strip():
            cal_file = 'dppp_predict_nocal.parset'
        else:
            cal_file = 'dppp_predict.parset'
            if self.settings.cal_mode == 'diagonal':
                cal_file = 'dppp_predict_diag.parset'

        DPPP.__init__(self, worker_settings, 'Predict', get_cal_config(cal_file), max_time=max_time)

    def get_parameters(self, msin):
        directions = self.sky_model.get_directions(msin, self.directions)
        if not directions:
            return None

        p = []
        p.append(f'msin.datacolumn={self.col_out}')
        p.append(f'msout.datacolumn={self.col_out}')
        p.append(f'predict.sourcedb={self.sky_model.get_sky_model(msin)}')
        if self.settings.parmdb.strip():
            p.append(f'predict.directions={self.get_directions_str(directions)}')
            p.append(f'predict.applycal.parmdb={msin}/{self.settings.parmdb}')
        else:
            p.append(f'predict.sources={self.get_directions_str_single_list(directions)}')

        return p


class Subtract(DPPP):

    def __init__(self, worker_settings, cal_settings, sky_model, col_in, col_out, directions='all', max_time=None):
        self.settings = cal_settings
        self.sky_model = sky_model
        self.col_in = col_in
        self.col_out = col_out
        self.directions = directions

        if not self.settings.parmdb.strip():
            cal_file = 'dppp_subtract_nocal.parset'
        else:
            cal_file = 'dppp_subtract.parset'
            if self.settings.cal_mode == 'diagonal':
                cal_file = 'dppp_subtract_diag.parset'

        DPPP.__init__(self, worker_settings, 'Subtract', get_cal_config(cal_file), max_time=max_time)

    def get_parameters(self, msin):
        directions = self.sky_model.get_directions(msin, self.directions)
        if not directions:
            return None

        p = []
        p.append(f'msin.datacolumn={self.col_in}')
        p.append(f'msout.datacolumn={self.col_out}')
        p.append(f'sub.sourcedb={self.sky_model.get_sky_model(msin)}')
        if self.settings.parmdb.strip():
            p.append(f'sub.directions={self.get_directions_str(directions)}')
            p.append(f'sub.applycal.parmdb={msin}/{self.settings.parmdb}')
        else:
            p.append(f'sub.sources={self.get_directions_str_single_list(directions)}')

        return p


class Peel(object):

    def __init__(self, sky_model):
        self.sky_model = sky_model

    def iterations(self, msins):
        n_pactchs = np.array([len(self.sky_model.get_patchs(msin, include_main=False)) for msin in msins])
        for i in np.arange(1, max(n_pactchs) + 1):
            yield i, np.array(msins)[n_pactchs >= i]


class PeelPreSubtract(Subtract):

    def __init__(self, peel_iter, worker_settings, cal_settings, sky_model, col_in='DATA', max_time=None):
        Subtract.__init__(self, worker_settings, cal_settings, sky_model, col_in, 'DATA_PEEL',
                          directions=slice(peel_iter, None), max_time=max_time)
        self.name = f'PeelPreSub {peel_iter}'
        self.peel_iter = peel_iter


class PeelCal(DDEcal):

    def __init__(self, peel_iter, worker_settings, cal_settings, sky_model, data_col='DATA_PEEL'):
        DDEcal.__init__(self, worker_settings, cal_settings, sky_model, data_col=data_col,
                        directions=peel_iter - 1)
        self.name = f'PeelCal {peel_iter}'
        self.peel_iter = peel_iter


class PeelPostSubtract(Subtract):

    def __init__(self, peel_iter, worker_settings, cal_settings, sky_model, data_col='DATA', max_time=None):
        Subtract.__init__(self, worker_settings, cal_settings, sky_model, data_col, data_col,
                          directions=peel_iter - 1, max_time=max_time)
        self.name = f'PeelPostSub {peel_iter}'
        self.peel_iter = peel_iter


class PeelPreSubtractPhaseShifted(PeelPreSubtract):

    def __init__(self, peel_iter, worker_settings, cal_settings, sky_model, col_in='DATA',
                 max_time=None, time_avg=4, freq_avg=1):
        PeelPreSubtract.__init__(self, peel_iter, worker_settings, cal_settings, sky_model,
                                 col_in=col_in, max_time=max_time)
        self.time_avg = time_avg
        self.freq_avg = freq_avg

    def get_out_file(self, msin):
        p = pathlib.Path(msin)
        return str(p.parent / ('tmp_' + p.name))

    def get_parameters(self, msin):
        coord = self.sky_model.get_patch_coord(msin, self.sky_model.get_patchs(msin)[self.peel_iter - 1])
        msout = self.get_out_file(msin)

        p = PeelPreSubtract.get_parameters(self, msin)
        p.append('steps=[sub,phaseshift,avg]')
        p.append('phaseshift.type=phaseshifter')
        p.append('avg.type=average')
        p.append(f'phaseshift.phasecenter=[{coord[0].deg}deg,{coord[1].deg}deg]')
        p.append(f'avg.timestep={self.time_avg}')
        p.append(f'avg.freqstep={self.freq_avg}')
        p.append(f'msout={msout}')
        p.append(f'msout.datacolumn=DATA')
        p.append(f'msout.overwrite=true')

        return p


class PeelPostSubtractPhaseShift(PeelPostSubtract):

    def __init__(self, peel_iter, worker_settings, cal_settings, sky_model, data_col='DATA', max_time=None,
                 coord_phase_back=None):
        PeelPostSubtract.__init__(self, peel_iter, worker_settings, cal_settings, sky_model,
                                  data_col=data_col, max_time=max_time)
        self.coord_phase_back = coord_phase_back

    def get_out_file(self, msin):
        p = pathlib.Path(msin)
        return str(p.parent / ('tmp_' + p.name))

    def get_parameters(self, msin):
        coord = self.sky_model.get_patch_coord(msin, self.sky_model.get_patchs(msin)[self.peel_iter - 1])
        msout = self.get_out_file(msin)

        if self.coord_phase_back:
            coord_phase_back = f'[{self.coord_phase_back.ra.deg}deg,{self.coord_phase_back.dec.deg}deg]'
        else:
            coord_phase_back = '[]'

        p = PeelPostSubtract.get_parameters(self, msin)
        p.append('steps=[phaseshift,sub,phaseshiftback]')
        p.append('phaseshift.type=phaseshifter')
        p.append('phaseshiftback.type=phaseshifter')
        p.append(f'phaseshift.phasecenter=[{coord[0].deg}deg,{coord[1].deg}deg]')
        p.append(f'phaseshiftback.phasecenter={coord_phase_back}')
        p.append(f'msout={msout}')
        p.append(f'msout.datacolumn=DATA')
        p.append(f'msout.overwrite=true')

        return p


class ApplyCal(DPPP):

    def __init__(self, worker_settings, cal_settings, col_in='DATA', col_out='CORRECTED_DATA', direction='Main'):
        self.col_in = col_in
        self.col_out = col_out
        self.direction = direction
        self.settings = cal_settings
        self.direction = direction
        dppp_file = 'dppp_applycal.parset'
        if self.settings.cal_mode == 'diagonal':
            dppp_file = 'dppp_applycal_diag.parset'

        DPPP.__init__(self, worker_settings, 'ApplyCal', get_cal_config(dppp_file))

    def get_parameters(self, msin):
        p = []
        p.append(f'apply.parmdb={msin}/{self.settings.parmdb}')
        p.append(f'apply.direction=[{self.direction}]')
        p.append(f'msin.datacolumn={self.col_in}')
        p.append(f'msout.datacolumn={self.col_out}')

        return p


class FlagPostCal(DPPP):

    def __init__(self, worker_settings, strategy='nenufar_1s1c'):
        if strategy == 'nenufar_1s1c':
            strategy = get_cal_config('LBAdefault_after_di.rfis')
        self.strategy = strategy
        DPPP.__init__(self, worker_settings, 'FlagPostCal', get_cal_config('dppp_flagger.parset'))

    def get_parameters(self, msin):
        p = []
        p.append(f'flag.strategy={self.strategy}')

        return p


class FlagBadStations(DPPP):

    def __init__(self, worker_settings, nsigma=5):
        self.nsigma = nsigma
        DPPP.__init__(self, worker_settings, 'FlagBadStations', get_cal_config('dppp_preflag.parset'))

    def get_parameters(self, msin):
        to_flag = flagutils.get_badstatsions(msin, self.nsigma)
        if len(to_flag) == 0:
            return None

        to_flag = ','.join([str(k) for k in to_flag])
        return [f'flag.baseline={to_flag}']


class FlagBaselines(DPPP):

    def __init__(self, worker_settings, baselines=''):
        self.baselines = baselines
        DPPP.__init__(self, worker_settings, 'FlagBaselines', get_cal_config('dppp_preflag.parset'))

    def get_baselines(self, msin):
        return self.baselines

    def get_parameters(self, msin):
        to_flag = self.get_baselines(msin)
        if len(to_flag) == 0:
            return None

        return [f'flag.baseline="{to_flag}"']


class FlagFreqs(DPPP):

    def __init__(self, worker_settings, fmhz_range):
        self.fmhz_range = fmhz_range
        DPPP.__init__(self, worker_settings, 'FlagFreqs', get_cal_config('dppp_preflag.parset'))

    def get_parameters(self, msin):
        return [f'flag.freqrange="{self.fmhz_range[0]}..{self.fmhz_range[1]} MHz"']


class FlagBadBaselines(FlagBaselines):

    def __init__(self, worker_settings, nsigma_stations=5, nsigma_baselines=8):
        self.nsigma_stations = nsigma_stations
        self.nsigma_baselines = nsigma_baselines
        FlagBaselines.__init__(self, worker_settings)

    def get_baselines(self, msin):
        return flagutils.get_badbaselines(msin, self.nsigma_stations, self.nsigma_baselines)


class SSINSFlagger(MultiCommands):

    def __init__(self, worker_settings, config='default', plot_dir=None, data_col='CORRECTED_DATA'):
        self.config = config
        self.plot_dir = plot_dir
        self.data_col = data_col
        MultiCommands.__init__(self, worker_settings, 'SSINS', 'flagtool ssins')

    def get_parameters(self, msin):
        p = [msin, f'--data_col={self.data_col}']
        if self.plot_dir is not None:
            plot_dir = os.path.join(msin, self.plot_dir)
            futils.mkdir(plot_dir)
            p.append(f'--plot_dir={plot_dir}')
        if self.config != 'default':
            p.append(f'--config={self.config}')
        return p


def main():
    pass

    # ret = TestMulti('test').run(msins)
    # print(ret)

    # AoQuality().run(msins)
    # PostProcessSolutions('instru_test.h5').run(msins)

    # CopyDataCol('DATA', 'SUBTRACTED_DATA').run(msins)

    # for i in range(3):
    #     parmdb_iter = f'instrument_peel_iter{i + 1}.h5'
    #     PeelCalibrate(i + 1, 'instrument_smooth.h5', parmdb_iter, sol_int_fct).run(msins)
    #     PostProcessSolutions(parmdb_iter).run(msins)
    #     PeelSubtract('SUBTRACTED_DATA', 'SUBTRACTED_DATA', parmdb_iter, i + 1).run(msins)

    # ApplyCal('SUBTRACTED_DATA', 'CORRECTED_DATA', 'instrument_smooth.h5', 'Main').run(msins)
    # AoQuality().run(msins)

    # FlagPostCal().run(msins)


if __name__ == '__main__':
    main()



import os
import time
import shutil

import numpy as np

from libpipe import futils

from . import skymodel, cal, flagutils, msutils


def get_all_tasks():
    d = {}
    for klass in AbstractTask.__subclasses__():
        if hasattr(klass, 'name'):
            d[klass.name] = klass
    return d


def get_all_tasks_descriptions():
    d = {}
    for klass in AbstractTask.__subclasses__():
        if hasattr(klass, 'name') and hasattr(klass, 'desc'):
            d[klass.name] = klass.desc
    return d


class AbstractTask(object):

    def __init__(self, s_task, s_worker, s_skymodel):
        self.s_task = s_task
        self.s_worker = s_worker
        self.s_skymodel = s_skymodel

    def run(self, msins):
        pass


class BuildSkyModel(AbstractTask):

    name = 'build_sky_model'
    desc = 'Build sky model'

    def __init__(self, s_task, s_worker, s_skymodel):
        AbstractTask.__init__(self, s_task, s_worker, s_skymodel)

    def copy_given_app_sky_model(self, msin):
        app_sky_model_name = self.s_skymodel.app_sky_model_name
        app_sky_model_file = self.s_skymodel.app_sky_model_file

        if not isinstance(app_sky_model_file, str):
            fmhz = msutils.get_ms_freqs(msin)[0].mean() * 1e-6
            fmhzs, filenames = list(zip(*sorted(app_sky_model_file.items())))
            fmhzs = [int(k) for k in fmhzs]
            i = np.where(np.array(fmhzs) - fmhz >= 0)[0][0]
            app_sky_model_file = filenames[i]

        app_sky_model_file = app_sky_model_file.replace('{MSIN}', msin)
        app_sky_model_file = os.path.abspath(app_sky_model_file)

        app_sky_model_out = cal.SkyModel(app_sky_model_name).get_sky_model_bbs(msin)

        print(f'Copy {app_sky_model_file} -> {app_sky_model_out}')
        futils.mkdir(os.path.dirname(app_sky_model_out))
        shutil.copyfile(app_sky_model_file, app_sky_model_out)

        sky_model = skymodel.lsmtool.load(app_sky_model_out)
        if 'Patch' not in sky_model.getColNames():
            print(f'Set default patch name Main for sky model {app_sky_model_out}')
            skymodel.set_patch_name(sky_model, 'Main')
            sky_model.write(app_sky_model_out, clobber=True)

    def run(self, msins):
        int_ateam_sky_model = self.s_skymodel.int_ateam_sky_model
        app_sky_model_name = self.s_skymodel.app_sky_model_name
        print(int_ateam_sky_model)
        if int_ateam_sky_model == 'lowres':
            int_ateam_sky_model = skymodel.get_lowres_ateam_skymodel_path()
        if self.s_task.add_ateam and not os.path.exists(int_ateam_sky_model):
            print('Intrinsic Ateam sky model not found. Can be either a file or "lowres" for the default one.')
            return []

        if self.s_skymodel.app_sky_model_file:
            for msin in msins:
                self.copy_given_app_sky_model(msin)
            if self.s_task.add_ateam:
                app_ateam_sky_model_name = 'app_sky_model_ateam'
                cal.MakeAppSkyModel(self.s_worker, app_ateam_sky_model_name, [int_ateam_sky_model],
                                    self.s_task).run(msins)

                cal.ConcatenateSkyModel(self.s_worker, app_sky_model_name,
                                        [app_sky_model_name, app_ateam_sky_model_name], self.s_task).run(msins)

        else:
            int_sky_model = self.s_skymodel.int_sky_model
            if int_sky_model in skymodel.sky_model_catalogs:
                int_sky_model = f'catalog_intrinsic_{int_sky_model}.skymodel'
                if not self.s_worker.dry_run:
                    skymodel.build_sky_model_ms(msins[0], self.s_task.min_flux, self.s_task.catalog_radius, int_sky_model,
                                                catalog=self.s_skymodel.int_sky_model)
            if not os.path.exists(int_sky_model):
                print('Intrinsic sky model not found. Can be either a file or one of the catalog: lcs165 or specfinf.')
                return []
            int_sky_models = [int_sky_model]

            if self.s_task.add_ateam:
                int_sky_models.append(int_ateam_sky_model)

            cal.MakeAppSkyModel(self.s_worker, app_sky_model_name, int_sky_models, self.s_task).run(msins)
        cal.MakeSourceDB(self.s_worker, app_sky_model_name).run(msins)

        return msins


class RestoreFlags(AbstractTask):

    name = 'restore_flags'
    desc = 'Restore or Backup flags'

    def __init__(self, s_task, s_worker, s_skymodel):
        AbstractTask.__init__(self, s_task, s_worker, s_skymodel)

    def run(self, msins):
        cal.RestoreOrBackupFlag(self.s_worker, self.s_task.flag_name).run(msins)

        return msins


class DDECal(AbstractTask):

    name = 'ddecal'
    desc = 'Direction dependent calibration'

    def __init__(self, s_task, s_worker, s_skymodel):
        AbstractTask.__init__(self, s_task, s_worker, s_skymodel)

    def run(self, msins):
        cal_settings = cal.CalSettings(**self.s_task.cal)
        sky_model = cal.SkyModel(self.s_skymodel.app_sky_model_name)

        if self.s_task.avg.time == 1 and self.s_task.avg.freq == 1:
            cal.DDEcal(self.s_worker, cal_settings, sky_model, data_col=self.s_task.col_in,
                       directions=self.s_task.directions).run(msins)
        else:
            mstemps = cal.DDEcalAvg(self.s_worker, cal_settings, sky_model, data_col=self.s_task.col_in,
                                    directions=self.s_task.directions, time_avg=self.s_task.avg.time,
                                    freq_avg=self.s_task.avg.freq).run(msins)
            futils.zip_rm(mstemps)

        if self.s_task.do_smooth_sol:
            futils.zip_copy(msins, msins, cal_settings.parmdb, filename_out=cal_settings.parmdb + '.bck')
            cal.SmoothSolutions(self.s_worker, cal_settings.parmdb, **self.s_task.smooth_sol).run(msins)

        if self.s_task.plot_sol:
            cal.PlotSolutions(self.s_worker, cal_settings.parmdb).run(msins)

        return msins


class Flagger(AbstractTask):

    name = 'flagger'
    desc = 'Pre/post calibration flagging'

    def __init__(self, s_task, s_worker, s_skymodel):
        AbstractTask.__init__(self, s_task, s_worker, s_skymodel)

    def run(self, msins):
        if self.s_task.do_aoflagger:
            cal.FlagPostCal(self.s_worker, self.s_task.aoflagger.strategy).run(msins)

        if self.s_task.do_baselinesflag:
            cal.FlagBaselines(self.s_worker, self.s_task.baselinesflag.baselines).run(msins)

        if self.s_task.do_flagfreq:
            cal.FlagFreqs(self.s_worker, self.s_task.flagfreq.fmhz_range).run(msins)

        if self.s_task.do_badbaselines:
            cal.AoQuality(self.s_worker).run(msins)
            cal.FlagBadBaselines(self.s_worker, **self.s_task.badbaselines).run(msins)

        if self.s_task.do_ssins:
            cal.SSINSFlagger(self.s_worker, config=self.s_task.ssins.seetings, plot_dir='flag_plot').run(msins)

        if self.s_task.do_scans_flagging:
            print('Start scans flagging ...')
            flagutils.flag_badscans(msins, data_col='CORRECTED_DATA', nsigma=self.s_task.scans_flagging.nsigma_scans)

        cal.AoQuality(self.s_worker).run(msins)

        return msins


class SmoothSolutions(AbstractTask):

    name = 'multims_smooth_sol'
    desc = 'Smooth Solutions over multiple MS'

    def __init__(self, s_task, s_worker, s_skymodel):
        AbstractTask.__init__(self, s_task, s_worker, s_skymodel)

    def run(self, msins):
        from .smoothsol import smooth_solutions

        smooth_solutions(msins, self.s_task.parmdb_in, self.s_task.parmdb_out, plot_dir=self.s_task.plot_dir)

        return msins


class ApplyCal(AbstractTask):

    name = 'apply_cal'
    desc = 'Apply calibration'

    def __init__(self, s_task, s_worker, s_skymodel):
        AbstractTask.__init__(self, s_task, s_worker, s_skymodel)

    def run(self, msins):
        cal_settings = cal.CalSettings(**self.s_task.cal)

        cal.ApplyCal(self.s_worker, cal_settings, self.s_task.col_in, self.s_task.col_out,
                     direction=self.s_task.direction).run(msins)

        return msins


class Subtract(AbstractTask):

    name = 'subtract'
    desc = 'Subtract sources'

    def __init__(self, s_task, s_worker, s_skymodel):
        AbstractTask.__init__(self, s_task, s_worker, s_skymodel)

    def run(self, msins):
        cal_settings = cal.CalSettings(**self.s_task.cal)
        sky_model = cal.SkyModel(self.s_skymodel.app_sky_model_name)
        cal.Subtract(self.s_worker, cal_settings, sky_model, self.s_task.col_in,
                     self.s_task.col_out, directions=self.s_task.directions).run(msins)

        return msins


class Predict(AbstractTask):

    name = 'predict'
    desc = 'predict sources'

    def __init__(self, s_task, s_worker, s_skymodel):
        AbstractTask.__init__(self, s_task, s_worker, s_skymodel)

    def run(self, msins):
        cal_settings = cal.CalSettings(**self.s_task.cal)
        sky_model = cal.SkyModel(self.s_skymodel.app_sky_model_name)
        cal.Predict(self.s_worker, cal_settings, sky_model, self.s_task.col_out, directions=self.s_task.directions).run(msins)

        return msins


class PeelCal(AbstractTask):

    name = 'peel'
    desc = 'Peel calibration'

    def __init__(self, s_task, s_worker, s_skymodel):
        AbstractTask.__init__(self, s_task, s_worker, s_skymodel)

    def run(self, msins):
        cal_settings_init = cal.CalSettings(**self.s_task.init)
        sky_model = cal.SkyModel(self.s_skymodel.app_sky_model_name)

        # Start with a new MS
        msouts = cal.CopyMS(self.s_worker, 'DATA', self.s_task.ms_postfix).run(msins)

        # Copy initial calibration table from .MS to _PEEL.MS
        futils.zip_copy(msins, msouts, cal_settings_init.parmdb)
        sky_model.copy(msins, msouts)

        for i, mspeel in cal.Peel(sky_model).iterations(msouts):
            cal_settings_iter = cal.CalSettings(**self.s_task.cal)
            cal_settings_iter.parmdb = f'instrument_peel_iter{i}.h5'

            if self.s_task.do_phase_shift:
                mstemps = cal.PeelPreSubtractPhaseShifted(i, self.s_worker, cal_settings_init, sky_model,
                                                          **self.s_task.phase_shift).run(mspeel)

                sky_model.copy(mspeel, mstemps)
                cal.PeelCal(i, self.s_worker, cal_settings_iter, sky_model, data_col='DATA').run(mstemps)

                if self.s_task.do_smooth_sol:
                    cal.SmoothSolutions(self.s_worker, cal_settings_iter.parmdb, **self.s_task.smooth_sol).run(mstemps)

                futils.zip_copy(mstemps, mspeel, cal_settings_iter.parmdb)
                futils.zip_rm(mstemps)

                mstemps = cal.PeelPostSubtractPhaseShift(i, self.s_worker, cal_settings_iter, sky_model).run(mspeel)

                time.sleep(1)

                futils.zip_rename_reg(mspeel, mstemps, 'table|^[A-Z]', invert=True)
                futils.zip_rm(mspeel)
                futils.zip_rename(mstemps, mspeel)
            else:
                cal.PeelPreSubtract(i, self.s_worker, cal_settings_init, sky_model).run(mspeel)
                cal.PeelCal(i, self.s_worker, cal_settings_iter, sky_model).run(mspeel)
                cal.PeelPostSubtract(i, self.s_worker, cal_settings_iter, sky_model).run(mspeel)

            cal.PlotSolutions(self.s_worker, cal_settings_iter.parmdb).run(mspeel)

        return msouts



import os

import numpy as np

import tables as h5_tables
from casacore import tables


class FlagManager(object):

    def __init__(self, time, freqs, flag):
        self.time = time
        self.freqs = freqs
        self.flag = flag

    @staticmethod
    def load_from_ms(ms_file):
        print(f'Loading flag from MS {ms_file} ...')
        with tables.table(ms_file, readonly=True) as t:
            flag = t.getcol('FLAG')
            time = t.getcol('TIME')
            freqs, _ = get_ms_freqs(ms_file)

        return FlagManager(time, freqs, flag)

    def save_to_ms(self, ms_file):
        print(f'Saving flag to MS {ms_file} ...')
        with tables.table(ms_file, readonly=False) as t:
            current_flag = t.getcol('FLAG')
            assert current_flag.shape == self.flag.shape
            t.putcol('FLAG', self.flag)

    @staticmethod
    def reset(ms_file):
        print(f'Reseting flag in MS {ms_file} ...')
        with tables.table(ms_file, readonly=False) as t:
            current_flag = t.getcol('FLAG')
            t.putcol('FLAG', np.zeros_like(current_flag, dtype=bool))

    @staticmethod
    def fully_flag(ms_file):
        print(f'Fully flagging MS {ms_file} ...')
        with tables.table(ms_file, readonly=False) as t:
            current_flag = t.getcol('FLAG')
            t.putcol('FLAG', np.ones_like(current_flag, dtype=bool))

    @staticmethod
    def load(h5_file):
        print(f'Loading flag file {h5_file} ...')
        with h5_tables.open_file(h5_file, 'r') as h5:
            flag = h5.root.flag.flag.read()
            freqs = h5.root.flag.freqs.read()
            time = h5.root.flag.time.read()

        return FlagManager(time, freqs, flag)

    def save(self, h5_file):
        print(f'Saving flag file {h5_file} ...')
        with h5_tables.open_file(h5_file, 'w') as h5:
            group = h5.create_group("/", 'flag', 'Flag cube')
            h5.create_array(group, 'flag', self.flag, "Flags")
            h5.create_array(group, 'freqs', self.freqs, "Frequencies (Hz)")
            h5.create_array(group, 'time', self.time, "Time")


class MsDataCube(object):

    def __init__(self, m_vis, m_vis_dt, time, weights, bu, ru, ant1, ant2, idx_baselines, freq, ant_name):
        self.data = m_vis
        self.data_dt = m_vis_dt
        self.time = time
        self.ru = ru
        self.bu = bu
        self.ant1 = ant1
        self.ant2 = ant2
        self.idx_baselines = idx_baselines
        self.freq = freq
        self.weights = weights
        self.ant_name = ant_name

    @staticmethod
    def load(ms_file, umin, umax, data_col, n_time_avg=1, flag_interstations=True, verbose=True, baselines='*&&',
             time_start_end=None):
        if verbose:
            print('Opening %s ...' % ms_file)
        if time_start_end is not None:
            start, end = time_start_end
            time_constraint = f'and TIME > {start} and TIME < {end}'
        else:
            time_constraint = ''
        with tables.taql(f"select from $ms_file where mscal.baseline($baselines) {time_constraint}") as t:
            uvw = t.getcol('UVW')
            data = t.getcol(data_col)
            flag = t.getcol('FLAG')
            time = t.getcol('TIME')
            a1 = t.getcol('ANTENNA1')
            a2 = t.getcol('ANTENNA2')

            weights = None
            if 'WEIGHT_SPECTRUM' in t.colnames() and t.iscelldefined('WEIGHT_SPECTRUM', 0):
                weights = t.getcol('WEIGHT_SPECTRUM')

        ant_name = tables.table(os.path.join(ms_file, 'ANTENNA')).getcol('NAME')

        baselines = a1 + 10000 * a2

        freq, _ = get_ms_freqs(ms_file)

        uu, vv, ww = uvw.T
        bu = np.sqrt(uu ** 2 + vv ** 2 + ww ** 2)
        ru = np.sqrt(uu ** 2 + vv ** 2)

        idx = (bu > umin) & (bu < umax)

        if is_lofar(ms_file) and flag_interstations:
            stat_name = np.array([k[:5] for k in ant_name])
            idx = idx & (stat_name[a1] != stat_name[a2])

        sel_baselines = np.unique(baselines[idx])

        if verbose:
            print('Selecting %s baselines between %s and %s meter' % (len(sel_baselines), umin, umax))

        idx_baselines = np.in1d(baselines, sel_baselines)
        n_time = data[idx_baselines].shape[0] // len(sel_baselines)

        vis = data[idx_baselines].reshape(n_time, len(sel_baselines), data.shape[1], data.shape[2])
        flag = flag[idx_baselines].reshape(n_time, len(sel_baselines), data.shape[1], data.shape[2])
        m_vis = np.ma.masked_array(vis, flag)

        time = time[idx_baselines].reshape(n_time, len(sel_baselines))[:, 0]
        if weights is not None:
            weights = weights[idx_baselines].reshape(n_time, -1, data.shape[1], data.shape[2])

        ant1 = a1[idx_baselines].reshape(n_time, len(sel_baselines))[0]
        ant2 = a2[idx_baselines].reshape(n_time, len(sel_baselines))[0]

        ru = ru[idx_baselines][:len(sel_baselines)]
        bu = bu[idx_baselines][:len(sel_baselines)]

        m_vis_dt = diff_consecutive(m_vis)

        if n_time_avg > 1:
            m_vis = mean_consecutive(m_vis, n=n_time_avg, axis=0)
            time = mean_consecutive(time, n=n_time_avg, axis=0)
            m_vis_dt = mean_consecutive(m_vis_dt, n=n_time_avg, axis=0)

        m_vis = m_vis.transpose((2, 0, 1, 3))
        m_vis_dt = m_vis_dt.transpose((2, 0, 1, 3))

        if weights is not None:
            weights = weights.transpose((2, 0, 1, 3))

        return MsDataCube(m_vis, m_vis_dt, time, weights, bu, ru, ant1, ant2, idx_baselines, freq, ant_name)

    def save_flag(self, ms, flag_new):
        with tables.table(ms, readonly=False) as t:
            flag_new = flag_new.transpose(1, 2, 0, 3)
            flag_new = flag_new.reshape(-1, flag_new.shape[2], flag_new.shape[3])

            flag = t.getcol('FLAG')
            flag[self.idx_baselines] = flag[self.idx_baselines] + flag_new

            t.putcol('FLAG', flag.astype(bool))


def slice_dim_idx(a, axis, s):
    idx = [slice(None) for k in range(a.ndim)]
    idx[axis] = s
    return tuple(idx)


def slice_dim(a, axis, s):
    return a[slice_dim_idx(a, axis, s)]


def mean_consecutive(a, n=2, axis=0, return_n=False):
    if isinstance(a, np.ma.MaskedArray):
        a_sum = np.add.reduceat(a.filled(0), np.arange(0, a.shape[axis], n), axis=axis)
        a_n_sum = np.add.reduceat((~a.mask).astype(int), np.arange(0, a.shape[axis], n), axis=axis)
        if return_n:
            return np.ma.array(a_sum / a_n_sum, mask=(a_n_sum == 0)), a_n_sum
        return np.ma.array(a_sum / a_n_sum, mask=(a_n_sum == 0))
    a_sum = np.add.reduceat(a, np.arange(0, a.shape[axis], n), axis=axis)
    return a_sum / float(n)


def diff_consecutive(a, axis=0):
    d = slice_dim(np.diff(a, axis=axis), axis, slice(None, None, 2))
    if isinstance(d, np.ma.MaskedArray) and d.mask.ndim == 0:
        d.mask = np.zeros_like(d)

    return d


def get_ms_freqs(ms_file):
    with tables.table(os.path.join(ms_file, 'SPECTRAL_WINDOW'), readonly=True, ack=False) as t_spec_win:
        freqs = t_spec_win.getcol('CHAN_FREQ').squeeze()
        chan_widths = t_spec_win.getcol('CHAN_WIDTH').squeeze()

    return freqs, chan_widths


def get_ms_time_interval(ms_file):
    with tables.table(ms_file) as t:
        return t.getcell('INTERVAL', 0)


def get_info_from_ms_files(files):
    t_start = []
    t_end = []
    t_total = 0

    for file in files:
        with tables.table(file, ack=False) as t:
            start = t.getcell('TIME', 0)
            end = t.getcell('TIME', t.nrows() - 1)
            t_int = t.getcell('INTERVAL', 0)
        t_start.append(start)
        t_end.append(end)
        t_total += end - start
        freqs, chan_widths = get_ms_freqs(file)
        n_chan = len(freqs)
        chan_width = chan_widths.mean()

    t_start = min(t_start)
    t_end = max(t_end)

    return {'start_time': t_start, 'end_time': t_end, 'int_time': t_int,
            'total_time': t_total, 'n_channels': n_chan, 'chan_width': chan_width}


def make_ant_matrix(ant1, ant2, m, a_max=54):
    m_map = np.ma.zeros((a_max, a_max)) * np.nan
    for a1, a2, i in zip(ant1, ant2, m):
        if (a1 < a_max) and (a2 < a_max):
            m_map[a1, a2] = i

    return m_map


def is_lofar(ms_file):
    return all(np.array(tables.table(os.path.join(ms_file, 'ANTENNA')).getcol('STATION')) == 'LOFAR')



import fnmatch
import itertools
from collections import defaultdict
import re

import toml

from libpipe import attrmap, worker


class DataHandler(object):

    def __init__(self, spectral_windows, data_level, obs_ids, n2_obs_ids, remote_hosts, l1_to_l2_config, nodes_distribution):
        self.spectral_windows = spectral_windows
        self.data_level = data_level
        self.obs_ids = obs_ids
        self.n2_obs_ids = n2_obs_ids
        self.remote_hosts = remote_hosts
        self.l1_to_l2_config = l1_to_l2_config
        self.nodes_distribution = nodes_distribution
        self.sw_index = dict(zip(self.get_spectral_windows(), range(len(self.get_spectral_windows()))))

    @staticmethod
    def from_file(filename):
        s = toml.load(filename, _dict=defaultdict)
        for n in ['obs_ids', 'spectral_windows', 'data_level_path']:
            assert n in s, f'{n} need to be defined in {filename}'

        return DataHandler(s['spectral_windows'], s['data_level_path'], s['obs_ids'], s.get('n2_obs_ids', dict()),
                           s.get('remote_hosts', dict()), s.get('l1_to_l2_config', dict()), 
                           s.get('nodes_distribution', dict()))

    def get_obs_ids(self, obs_id_match=None, include_n2_obs_ids=True):
        all_obs_ids = list(self.obs_ids.keys() )
        if include_n2_obs_ids:
            all_obs_ids += list(self.n2_obs_ids.keys())
        if obs_id_match is None:
            return all_obs_ids
        return attrmap.OrderedSet(obs_id for k in obs_id_match.split(',') for obs_id in fnmatch.filter(all_obs_ids, k))

    def is_n2_obs_id(self, obs_id):
        return obs_id in self.n2_obs_ids.keys()

    def get_nodes_distribution(self):
        return self.nodes_distribution

    def get_obs_ids_and_spectral_windows(self, obs_id_sw_match):
        if ":" in obs_id_sw_match:
            obs_id_sw_match, sws_str = obs_id_sw_match.split(':')
            sws = fnmatch.filter(self.get_spectral_windows(), sws_str)
        else:
            sws = self.get_spectral_windows()

        return self.get_obs_ids(obs_id_sw_match), sws

    def get_sbs(self, sw):
        sb_l, sb_r = self.spectral_windows[sw]
        return range(int(sb_l), int(sb_r) + 1)

    def get_all_hosts(self):
        return [*set(n for l in self.obs_ids.values() for n in l)]

    def get_remote_hosts(self):
        return self.remote_hosts.keys()

    def get_remote_host(self, remote_host):
        assert remote_host in self.remote_hosts, f'{remote_host} needs to be defined'
        return self.remote_hosts[remote_host]['host']

    def get_remote_level(self, remote_host):
        assert remote_host in self.remote_hosts, f'{remote_host} needs to be defined'
        return self.remote_hosts[remote_host]['level']

    def get_remote_password_file(self, remote_host):
        assert remote_host in self.remote_hosts, f'{remote_host} needs to be defined'

        return self.remote_hosts[remote_host]['password_file']

    def get_remote_data_path(self, remote_host):
        assert remote_host in self.remote_hosts, f'{remote_host} needs to be defined'

        return self.remote_hosts[remote_host]['data_path']

    def get_l1_to_l2_config(self, l2_level):
        assert l2_level in self.l1_to_l2_config, f'{l2_level} needs to be defined'

        return self.l1_to_l2_config[l2_level]['dppp_config']

    def get_levels(self):
        return self.data_level.keys()

    def get_spectral_windows(self):
        return self.spectral_windows.keys()

    def get_node(self, obs_id, sw):
        assert not self.is_n2_obs_id(obs_id)
        return self.obs_ids[obs_id][self.sw_index[sw]]

    def get_n2_nodes(self, n2_obs_id):
        nodes = self.n2_obs_ids[n2_obs_id]['nodes']
        if isinstance(nodes, str):
            nodes = sorted(worker.get_hosts(nodes))
        return nodes

    def get_n1_obs_ids(self, n2_obs_id):
        return self.n2_obs_ids[n2_obs_id]['n1_obs_ids']

    def get_dir_path(self, obs_id, level, sw):
        node = self.get_node(obs_id, sw)
        return f'{self.data_level[level].replace("%NODE%", node)}/{obs_id}/'

    def get_ms_path(self, obs_id, level, sw):
        if self.is_n2_obs_id(obs_id):
            if level.lower().startswith('l1'):
                return list(itertools.chain.from_iterable(self.get_ms_path(n1_obs_id, level, sw) for n1_obs_id in self.get_n1_obs_ids(obs_id)))

            mss = []
            for i, node in enumerate(self.get_n2_nodes(obs_id)):
                mss.append(f'{self.data_level[level].replace("%NODE%", node)}/{obs_id}/{sw}_T{i:03}.MS')
            return mss
        else:
            dir_path = self.get_dir_path(obs_id, level, sw)

            if level.lower().startswith('l1'):
                return [f'{dir_path}/SB{sb}.MS' for sb in self.get_sbs(sw)]

            return [f'{dir_path}/{sw}.MS']

    def get_all_ms_path(self, obs_ids, levels, sws):
        for obs_id in obs_ids:
            for level in levels:
                for sw in sws:
                    yield from self.get_ms_path(obs_id, level, sw)



# Handle setting
#
# Author: F. Mertens

import os
import re

from libpipe.settings import BaseSettings


TEMPLATE_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'templates')


class Settings(BaseSettings):

    DEFAULT_SETTINGS = os.path.join(TEMPLATE_DIR, 'default_settings.toml')

    def __init__(self, file, d):
        BaseSettings.__init__(self, file, d)

    def get_target_host(self, in_file):
        host = None
        if self.worker.run_on_file_host and self.worker.run_on_file_host_pattern:
            r = re.search(self.worker.run_on_file_host_pattern, in_file)
            if r is not None:
                host = r.group(1)
        return host


class ImgSettings(Settings):

    DEFAULT_SETTINGS = os.path.join(TEMPLATE_DIR, 'default_img_settings.toml')

    def __init__(self, file, d):
        BaseSettings.__init__(self, file, d)



#!/usr/bin/env python

import os
from multiprocessing import Pool

import click

import numpy as np

import losoto.h5parm

import matplotlib as mpl
mpl.use('Agg')

import matplotlib.pyplot as plt

from nenucal import __version__

# Lazy loaded:
# - astropy.stats and  astropy.convolution in astro_convolve


mpl.rcParams['image.cmap'] = 'Spectral_r'
mpl.rcParams['image.origin'] = 'lower'
mpl.rcParams['image.interpolation'] = 'nearest'
mpl.rcParams['axes.grid'] = True

t_file = click.Path(exists=True, dir_okay=False)


class GainSol(object):

    def __init__(self, time, freqs, ant, directions, pol, amp, phase):
        self.time = time
        self.freqs = freqs
        self.ant = ant
        self.dir = directions
        self.pol = pol
        self.amp = amp
        self.phase = phase
        self.d = self.amp * np.exp(1j * self.phase)


@click.group()
@click.version_option(__version__)
def main():
    ''' DPPP gains solution utilities ...'''


def ctoap(r, i):
    c = r + 1j * i
    return abs(c), np.angle(c)


def aptoc(amp, phase):
    c = amp * np.exp(1j * phase)
    return c.real, c.imag


def get_next_odd(f):
    f = int(np.ceil(f))
    return f + 1 if f % 2 == 0 else f


def clip_and_smooth(amp, phase, x_stddev, y_stddev, sigma_clip=5):
    import astropy.stats as astats
    from astropy.convolution import Gaussian2DKernel
    from astropy.convolution import convolve

    xs = get_next_odd(np.clip(10 * x_stddev, 25, 100))
    ys = get_next_odd(np.clip(10 * y_stddev, 25, 100))

    kernel = Gaussian2DKernel(x_stddev=x_stddev, y_stddev=y_stddev, x_size=xs, y_size=ys)
    c_phase = np.exp(1j * phase)

    amp = astats.sigma_clip(amp, sigma=sigma_clip, maxiters=10)
    c_phase.mask = amp.mask

    amp = convolve(amp, kernel, boundary='extend')
    p_real = convolve(c_phase.real, kernel, boundary='extend')
    p_imag = convolve(c_phase.imag, kernel, boundary='extend')

    return amp, np.angle(p_real + 1j * p_imag)


def smooth_sol(sol, fwhm_time, main_fwhm_time, fwhm_freq, main_fwhm_freq, main_name='main', sigma_clip=5):
    dx_min = (sol.time[1] - sol.time[0]) / 60.
    dx_mhz = (sol.freqs[1] - sol.freqs[0]) * 1e-6

    s_amp = np.zeros_like(sol.amp)
    s_phase = np.zeros_like(sol.phase)

    for i_c in range(len(sol.dir)):
        if sol.dir[i_c].strip('[] ').lower() == main_name:
            c_fwhm_time = main_fwhm_time
            c_fwhm_freq = main_fwhm_freq
        else:
            c_fwhm_time = fwhm_time
            c_fwhm_freq = fwhm_freq
        stddev_time = c_fwhm_time / dx_min / 2.3
        stddev_freqs = c_fwhm_freq / dx_mhz / 2.3
        print(f'Smoothing {sol.dir[i_c]} with a Gaussian kernel of FWHM {c_fwhm_time} min and {c_fwhm_freq} MHz')

        for i_s in range(len(sol.ant)):
            for i_p in range(len(sol.pol)):
                a, p = clip_and_smooth(sol.amp[:, :, i_s, i_c, i_p], sol.phase[:, :, i_s, i_c, i_p], 
                                      stddev_freqs, stddev_time, sigma_clip=sigma_clip)
                s_amp[:, :, i_s, i_c, i_p] = a
                s_phase[:, :, i_s, i_c, i_p] = p

    return GainSol(sol.time, sol.freqs, sol.ant, sol.dir, sol.amp, s_amp, s_phase)


def open_sol(file_h5):
    # amp: time, freqs, antm, dir, pol
    sol_file = None
    try:
        sol_file = losoto.h5parm.h5parm(file_h5)
        solset = sol_file.getSolsets()[0]
        soltab, soltab_phase = solset.getSoltabs(useCache=True)

        ant = soltab.getAxisValues('ant')
        directions = soltab.getAxisValues('dir')
        time = soltab.getAxisValues('time')
        pol = soltab.getAxisValues('pol')

        freqs = soltab.getAxisValues('freq')

        weight = soltab.getValues(weight=True)[0].astype(bool)
        amp = np.ma.array(soltab.getValues(weight=False)[0], mask=~weight)
        phase = np.ma.array(soltab_phase.getValues(weight=False)[0], mask=~weight)

        if directions is None:
            amp = amp[:, :, :, None, :]
            phase = phase[:, :, :, None, :]
            directions = ['di']
    finally:
        if sol_file is not None:
            sol_file.close()

    return GainSol(time, freqs, ant, directions, pol, amp, phase)


def save_sol(file_h5, sol):
    sol_file = None
    try:
        print('Saving solutions ...')
        sol_file = losoto.h5parm.h5parm(file_h5, readonly=False)
        solset = sol_file.getSolsets()[0]
        soltab, soltab_phase = solset.getSoltabs()
        
        soltab.setValues(sol.amp)
        soltab_phase.setValues(sol.phase)
    finally:
        if sol_file is not None:
            sol_file.close()
        print('Done')


def plot_sol(sol, dir, pol, data_type, filename):
    if data_type == 'Amplitude':
        v = sol.amp[:, :, :, dir, pol]
    elif data_type == 'Phase':
        v = sol.phase[:, :, :, dir, pol]
    else:
        print(f'Error: data type {data_type} unknown')
        return

    vmax = np.nanquantile(v[~v.mask & ~np.isnan(v) & (v != 0)], 0.999)
    vmin = np.nanquantile(v[~v.mask & ~np.isnan(v) & (v != 0)], 0.001)
    extent = [0, len(sol.time), sol.freqs.min() * 1e-6, sol.freqs.max() * 1e-6]

    n = v.shape[2]
    ncols, nrows = int(np.ceil(np.sqrt(n))), int(np.ceil(n / np.ceil(np.sqrt(n))))

    fig, axs = plt.subplots(ncols=ncols, nrows=nrows, sharey=True, figsize=(1 + 2 * ncols, 1 + 1.5 * nrows),
                            sharex=True)

    im = None

    for i, ax in zip(range(v.shape[2]), axs.flatten()):
        if v.shape[0] > 1 and v.shape[1] > 1:
            im = ax.imshow(v[:, :, i].T, aspect='auto', vmax=vmax, vmin=vmin, extent=extent)
        elif v.shape[0] == 1:
            ax.plot(sol.freqs * 1e-6, v[0, :, i].T)
        elif v.shape[1] == 1:
            ax.plot(v[:, 0, i].T)
        ax.text(0.025, 0.975, sol.ant[i], transform=ax.transAxes, fontsize=11, va='top')

    ylabel = ''
    xlabel = ''

    if v.shape[0] > 1 and v.shape[1] > 1 and im is not None:
        cax = fig.add_axes([0.6, 1.04, 0.39, 0.02])
        cax.set_xlabel(data_type)
        fig.colorbar(im, cax=cax, orientation='horizontal')
        xlabel = 'Time index'
        ylabel = 'Frequency [Mhz]'
    elif v.shape[0] == 1:
        xlabel = 'Frequency [MHz]'
        ylabel = data_type
    elif v.shape[1] == 1:
        xlabel = 'Time index'
        ylabel = data_type

    for ax in axs[:, 0]:
        ax.set_ylabel(ylabel)
    for ax in axs[-1, :]:
        ax.set_xlabel(xlabel)

    fig.tight_layout(pad=0)
    fig.savefig(filename, dpi=120, bbox_inches="tight")


@main.command('smooth')
@click.argument('sols', nargs=-1, type=t_file)
@click.option('--fwhm_time', help='Time coherence scale (min)', type=float, default=16)
@click.option('--fwhm_freq', help='Freq coherence scale (MHz)', type=float, default=2)
@click.option('--main_fwhm_time', help='Time coherence scale (min) for Main direction', type=float, default=20)
@click.option('--main_fwhm_freq', help='Freq coherence scale (MHz) for Main direction', type=float, default=4)
@click.option('--clip_nsigma', help='Clip solution above NSIGMA', type=float, default=4)
@click.option('--main_name', help='Name of the main direction', type=str, default='main')
def smooth(sols, fwhm_time, fwhm_freq, main_fwhm_time, main_fwhm_freq, clip_nsigma, main_name):
    ''' Smooth solutions with a Gaussian kernel'''
    for sol_file in sols:
        sol = open_sol(sol_file)
        s_sol = smooth_sol(sol, fwhm_time, main_fwhm_time, fwhm_freq, main_fwhm_freq, main_name=main_name, sigma_clip=clip_nsigma)
        save_sol(sol_file, s_sol)


@main.command('plot')
@click.argument('sols', nargs=-1, type=t_file)
@click.option('--plot_dir', help='Plot directory', type=str, default='sol_plots')
@click.option('--n_cpu', help='Number of CPU to use', type=int, default=4)
def plot(sols, plot_dir, n_cpu):
    ''' Plot solutions of the h5 files SOLS '''
    for sol_file in sols:
        sol = open_sol(sol_file)

        with Pool(n_cpu) as pool:
            for data_type in ['Amplitude', 'Phase']:
                for dir in range(len(sol.dir)):
                    for pol in range(len(sol.pol)):
                        filename = f'{data_type}_dir{sol.dir[dir]}_pol{sol.pol[pol]}.png'
                        path = os.path.join(os.path.dirname(sol_file), plot_dir, filename)

                        if not os.path.exists(os.path.dirname(path)):
                            os.makedirs(os.path.dirname(path))

                        # plot_sol(sol, dir, pol, data_type, path)
                        pool.apply_async(plot_sol, [sol, dir, pol, data_type, path])
            pool.close()
            pool.join()


if __name__ == '__main__':
    main()



#!/usr/bin/env python

import os

import numpy as np

from scipy.interpolate import interp1d
from scipy.interpolate import SmoothBivariateSpline

import GPy

import astropy.stats as astats
import astropy.time as atime

from casacore import tables as pt

import losoto
import losoto.h5parm

import matplotlib as mpl
mpl.use('Agg')

import matplotlib.pyplot as plt

mpl.rcParams['image.cmap'] = 'Spectral_r'
mpl.rcParams['image.origin'] = 'lower'
mpl.rcParams['image.interpolation'] = 'nearest'
mpl.rcParams['axes.grid'] = True
mpl.rcParams['savefig.dpi'] = 150


def ctoap(r, i):
    c = r + 1j * i
    return abs(c), np.angle(c)


def aptoc(amp, phase):
    c = amp * np.exp(1j * phase)
    return c.real, c.imag


class Gains(object):

    def __init__(self, times, freqs, amp, phase, ant, pol, direction):
        self.times = times
        self.freqs = freqs
        self.amp = amp
        self.phase = phase
        self.ant = ant
        self.pol = pol
        self.direction = direction
        self.freqs_norm = freqs * 1e-6
        self.times_norm = times / 3600.

    @staticmethod
    def norm_dir(direction):
        return direction.strip('[]')

    @staticmethod
    def norm_dirs(directions):
        return np.array([Gains.norm_dir(k) for k in directions])

    @staticmethod
    def from_h5_files(h5_files, direction='Main'):
        amp_gains = []
        phase_gains = []
        times = []
        freqs = []
        ant = None
        pol = None

        for file_h5 in h5_files:
            sol_file = losoto.h5parm.h5parm(file_h5)
            solset = sol_file.getSolsets()[0]
            soltab, soltab_phase = solset.getSoltabs(useCache=True)

            idx_main = np.where(Gains.norm_dirs(soltab.getAxisValues('dir')) == Gains.norm_dir(direction))[0]
            if len(idx_main) == 0:
                continue
            idx_main = idx_main[0]

            ant = soltab.getAxisValues('ant')
            pol = soltab.getAxisValues('pol')

            freq = soltab.getAxisValues('freq')
            i_freq = [0, int(len(freq) / 2), len(freq) - 1]

            weight = soltab.getValues(weight=True)[0].astype(bool)
            amp = np.ma.mean(np.ma.array(soltab.getValues(weight=False)[0], mask=~weight), axis=0)[i_freq]
            phase = np.ma.mean(np.ma.array(soltab_phase.getValues(weight=False)[0], mask=~weight), axis=0)[i_freq]

            amp = amp[:, :, idx_main]
            phase = phase[:, :, idx_main]

            time = np.ma.array(soltab.getAxisValues('time'),
                               mask=~weight[:, :, :, idx_main, 0].mean(axis=(1, 2)).astype(bool))

            sol_file.close()

            if time.mask.sum() == time.size:
                print("Skipping", file_h5)
                continue

            time = [time.mean()] * len(i_freq)
            freq = freq[i_freq]

            amp_gains.extend(amp)
            phase_gains.extend(phase)
            times.extend(time)
            freqs.extend(freq)

        amp_gains = np.ma.array(amp_gains)
        phase_gains = np.ma.array(phase_gains)
        times = np.array(times)
        freqs = np.array(freqs)

        return Gains(times, freqs, amp_gains, phase_gains, ant, pol, direction)

    def build_gain_model(self, times_pre, freqs_pre, time_coh_scale_hour=20,
                         freq_coh_scale_mhz=10, method='gpr', var_snr=0.01):
        assert method in ['spline', 'gpr']

        kern = GPy.kern.RBF(2, ARD=True, lengthscale=[time_coh_scale_hour, freq_coh_scale_mhz])
        var_noise = var_snr * np.sqrt(0.5) * np.ma.median(self.amp)

        times_pre_norm = times_pre / 3600.
        freqs_pre_norm = freqs_pre * 1e-6
        X_pre = np.vstack([times_pre_norm.flatten(), freqs_pre_norm.flatten()]).T

        pre_amp = np.zeros((times_pre.size, self.amp.shape[1], self.amp.shape[2]))
        pre_phase = np.zeros((times_pre.size, self.amp.shape[1], self.amp.shape[2]))

        for pol_id in range(self.amp.shape[2]):
            for i in range(self.amp.shape[1]):
                m = ~self.amp[:, i, pol_id].mask
                if not np.all(m):
                    continue
                if np.all(m):
                    m = slice(None)
                freqs_norm = self.freqs[m] * 1e-6
                times_norm = self.times[m] / 3600.

                X = np.vstack((times_norm, freqs_norm)).T
                Y = np.dstack(aptoc(self.amp[m, i, pol_id], self.phase[m, i, pol_id])).reshape((X.shape[0], -1))

                mask_k_sigma = astats.sigma_clip(Y, sigma=4).mask.sum(axis=1) > 0
                X = X[~mask_k_sigma]
                Y = Y[~mask_k_sigma]

                Y_detrend = np.array([k - np.poly1d(np.polyfit(X[:, 1], Y[:, 0], 2))(X[:, 1]) for k in Y.T]).T
                mask_k_sigma = astats.sigma_clip(Y_detrend, sigma=4, axis=0).mask.sum(axis=1) > 0
                X = X[~mask_k_sigma]
                Y = Y[~mask_k_sigma]

                if method == 'gpr':
                    model = GPy.models.GPRegression(X, Y, kern, noise_var=var_noise)
                    Y_pre = model.predict(X_pre)[0].reshape(-1, 2)
                    Y_pre_amp, Y_pre_phase = ctoap(Y_pre[:, 0], Y_pre[:, 1])
                elif method == 'spline':
                    Y_pre_r = SmoothBivariateSpline(X[:, 0], X[:, 1], Y[:, 0], kx=1, ky=2)(
                        X_pre[:, 0], X_pre[:, 1], grid=False)
                    Y_pre_i = SmoothBivariateSpline(X[:, 0], X[:, 1], Y[:, 1], kx=1, ky=2)(
                        X_pre[:, 0], X_pre[:, 1], grid=False)
                    Y_pre_amp, Y_pre_phase = ctoap(Y_pre_r, Y_pre_i)

                pre_amp[:, i, pol_id] = Y_pre_amp
                pre_phase[:, i, pol_id] = Y_pre_phase

        return Gains(times_pre, freqs_pre, pre_amp, pre_phase, self.ant, self.pol, self.direction)

    def prepare_h5(self):
        time = np.unique(self.times)
        freq = np.unique(self.freqs)
        assert self.amp.shape[0] == len(time) * len(freq)

        amps = self.amp.reshape((len(freq), len(time), self.amp.shape[1], 1, self.amp.shape[2]))
        phases = self.phase.reshape((len(freq), len(time), self.amp.shape[1], 1, self.amp.shape[2]))

        amps = amps.transpose(1, 0, 2, 3, 4)
        phases = phases.transpose(1, 0, 2, 3, 4)

        return time, freq, amps, phases

    def save_to_h5(self, h5_file):
        time, freq, amps, phases = self.prepare_h5()
        weights = np.ones_like(amps)

        axis_vals = [time, freq, self.ant, ['[%s]' % self.direction], self.pol]
        axes_name = ['time', 'freq', 'ant', 'dir', 'pol']

        if os.path.exists(h5_file):
            os.remove(h5_file)

        los_h5 = losoto.h5parm.h5parm(h5_file, readonly=False)
        los_h5.makeSolset('sol000')
        sol_set = los_h5.getSolset('sol000')
        sol_set.makeSoltab('amplitude', 'amplitude000', axesNames=axes_name, axesVals=axis_vals,
                           vals=amps, weights=weights)
        sol_set.makeSoltab('phase', 'phase000', axesNames=axes_name, axesVals=axis_vals, vals=phases, weights=weights)
        los_h5.close()


class MultiGain(object):

    def __init__(self):
        self.dirs = dict()

    def add(self, direction, gains):
        self.dirs[direction] = gains

    @staticmethod
    def get_all_directions(h5_files):
        directions = set()
        for file_h5 in h5_files:
            sol_file = losoto.h5parm.h5parm(file_h5)
            solset = sol_file.getSolsets()[0]
            directions = directions.union(set(solset.getSoltabs()[0].getAxisValues('dir')))
            sol_file.close()
        return Gains.norm_dirs(directions)

    @staticmethod
    def from_h5_files(h5_files, directions=None):
        all_dirs = MultiGain.get_all_directions(h5_files)
        multi_gains = MultiGain()

        for direction in all_dirs:
            if directions is not None and direction not in directions:
                continue
            gains = Gains.from_h5_files(h5_files, direction=direction)
            multi_gains.add(direction, gains)

        return multi_gains

    def build_gain_model(self, times_pre, freqs_pre, time_coh_scale_hour, freq_coh_scale_mhz,
                         var_snr=0.01, method='gpr', directions=None):
        multi_model_gains = MultiGain()

        for direction, gains in self.dirs.items():
            if directions is not None and direction not in directions:
                continue
            model_gains = gains.build_gain_model(times_pre, freqs_pre, time_coh_scale_hour[direction],
                                                 freq_coh_scale_mhz[direction], method=method, var_snr=var_snr)
            multi_model_gains.add(direction, model_gains)

        return multi_model_gains

    def save_to_h5(self, h5_file):
        all_amps = []
        all_phases = []

        assert len(self.dirs) > 0

        for direction, gains in self.dirs.items():
            time, freq, amps, phases = gains.prepare_h5()
            all_amps.append(amps)
            all_phases.append(phases)

        amps = np.concatenate(all_amps, axis=3)
        phases = np.concatenate(all_phases, axis=3)
        weights = np.ones_like(amps)

        axis_vals = [time, freq, gains.ant, ['[%s]' % k for k in self.dirs.keys()], gains.pol]
        axes_name = ['time', 'freq', 'ant', 'dir', 'pol']

        if os.path.exists(h5_file):
            os.remove(h5_file)

        los_h5 = losoto.h5parm.h5parm(h5_file, readonly=False)
        los_h5.makeSolset('sol000')
        sol_set = los_h5.getSolset('sol000')
        sol_set.makeSoltab('amplitude', 'amplitude000', axesNames=axes_name, axesVals=axis_vals,
                           vals=amps, weights=weights)
        sol_set.makeSoltab('phase', 'phase000', axesNames=axes_name, axesVals=axis_vals, vals=phases, weights=weights)
        los_h5.close()


def copy_weights(in_h5, out_h5, direction=None):
    sol_file = losoto.h5parm.h5parm(in_h5)
    solset = sol_file.getSolsets()[0]
    soltab = solset.getSoltabs()[0]
    weight = soltab.getValues(weight=True, retAxesVals=False)
    time_sol = soltab.getAxisValues('time')
    freq_sol = soltab.getAxisValues('freq')
    if direction is not None:
        idx_main = np.where(soltab.getAxisValues('dir') == '[%s]' % direction)[0][0]
        weight = weight[:, :, :, idx_main, :][:, :, :, None, :]
    sol_file.close()

    out_sol_file = losoto.h5parm.h5parm(out_h5, readonly=False)
    solset = out_sol_file.getSolsets()[0]
    soltab_amp, soltab_phase = solset.getSoltabs()
    out_time_sol = soltab_amp.getAxisValues('time')
    out_freq_sol = soltab_amp.getAxisValues('freq')

    weight_out = weight.copy()
    if weight.shape[0] > 1:
        weight_out = interp1d(time_sol, weight_out, kind='nearest', axis=0)(out_time_sol)
    else:
        weight_out = np.repeat(weight_out, len(out_time_sol), axis=0)

    if weight.shape[1] > 1:
        weight_out = interp1d(freq_sol, weight_out, kind='nearest', axis=1)(out_freq_sol)
    else:
        weight_out = np.repeat(weight_out, len(out_freq_sol), axis=1)

    soltab_amp.setValues(weight_out, weight=True)
    soltab_phase.setValues(weight_out, weight=True)

    out_sol_file.close()


def get_directions(h5_file):
    sol_file = losoto.h5parm.h5parm(h5_file)
    solset = sol_file.getSolsets()[0]
    soltab = solset.getSoltabs()[0]
    directions = soltab.getAxisValues('dir')
    sol_file.close()

    return Gains.norm_dirs(directions)


def plot_smooth_gains(name, gains, plot_dir='.', time_coh_scale_hour=20,
                      freq_coh_scale_mhz=10, method='gpr'):
    times_pre = np.linspace(np.nanmin(gains.times) - 1e3, np.nanmax(gains.times) + 1e3, num=20)
    freqs_pre = np.linspace(np.nanmin(gains.freqs) - 1e6, np.nanmax(gains.freqs) + 1e6, num=20)
    times_pre, freqs_pre = np.meshgrid(times_pre, freqs_pre)

    gain_smooth = gains.build_gain_model(times_pre, freqs_pre, time_coh_scale_hour,
                                         freq_coh_scale_mhz, method)

    if not os.path.isdir(plot_dir):
        os.makedirs(plot_dir)

    for type in ['amp', 'phase']:
        extent = [gain_smooth.times_norm.min(), gain_smooth.times_norm.max(),
                  gain_smooth.freqs_norm.min(), gain_smooth.freqs_norm.max()]

        for pol_id, pol in enumerate(gains.pol):
            if type == 'amp':
                vmin = np.min(gain_smooth.amp[:, :, pol_id])
                vmax = np.max(gain_smooth.amp[:, :, pol_id])
                d_m = gain_smooth.amp
                d = gains.amp
            else:
                vmin = - np.pi
                vmax = np.pi
                d_m = gain_smooth.phase
                d = gains.phase

            fig, axs = plt.subplots(ncols=7, nrows=8, sharex=True, sharey=True,
                                    constrained_layout=False, figsize=(12, 14))
            axs = axs.flatten()

            for i in range(56):
                im = axs[i].imshow(d_m[:, i, pol_id].reshape(len(times_pre), len(freqs_pre)),
                                   aspect='auto', extent=extent, vmin=vmin, vmax=vmax)
                axs[i].scatter(gains.times_norm, gains.freqs_norm, c=d[:, i, pol_id].flatten(),
                               vmin=vmin, vmax=vmax, edgecolors='gray')
            #     axs[i].set_xlim(0, times_pre.max())
                axs[i].text(0.05, 0.97, str(i), transform=axs[i].transAxes, va='top', ha='left')

            fig.tight_layout(pad=2, h_pad=0, w_pad=0)
            axs[21].set_ylabel('Frequency [MHz]')
            axs[52].set_xlabel('Time [h]')
            fig.colorbar(im, ax=axs.tolist(), use_gridspec=True, fraction=0.02, pad=0.005, aspect=35)

            fig.savefig(f'{plot_dir}/smooth_gain_{name}_{type}_{pol}.png')


def get_time_freq(ms_file):
    with pt.table(ms_file) as t:
        time = np.unique(t.getcol('TIME'))
    with pt.table(ms_file + '/SPECTRAL_WINDOW') as t:
        freq = t.getcol('CHAN_FREQ').squeeze()

    return time, freq


def smooth_solutions(ms_files, parmdb_in, parmdb_out, time_scale=30, freq_scale=15,
                     time_scale_ateam=0.5, freq_scale_ateam=4, plot_dir='.', method='gpr'):
    h5_files = [os.path.join(f, parmdb_in) for f in ms_files]

    time, freq = get_time_freq(ms_files[0])
    observing_time = atime.Time(time.mean() / 3600. / 24., scale='utc', format='mjd')
    name = f'{observing_time.strftime("%Y%m%d")}_{int(freq.mean() * 1e-6)}MHz'

    print('Loading h5 files ...')
    gains = MultiGain.from_h5_files(h5_files)

    time_scale_dir = {}
    freq_scale_dir = {}
    for direction in gains.dirs.keys():
        if direction == 'Main':
            time_scale_dir[direction] = time_scale
            freq_scale_dir[direction] = freq_scale
        else:
            time_scale_dir[direction] = time_scale_ateam
            freq_scale_dir[direction] = freq_scale_ateam

    print('Plotting smooth gains ...')
    for direction, dir_gains in gains.dirs.items():
        plot_smooth_gains(name + f'_{direction}', dir_gains, plot_dir=plot_dir,
                          time_coh_scale_hour=time_scale_dir[direction],
                          freq_coh_scale_mhz=freq_scale_dir[direction], method=method)

    print('Building smooth gains version ...')
    for in_h5, ms_file in zip(h5_files, ms_files):
        directions = get_directions(in_h5)
        out_h5 = os.path.join(ms_file, parmdb_out)

        time, freq = get_time_freq(ms_file)
        time, freq = np.meshgrid(time, freq)

        gain_smooth = gains.build_gain_model(time, freq, time_coh_scale_hour=time_scale_dir,
                                             freq_coh_scale_mhz=freq_scale_dir,
                                             method=method, directions=directions)

        print(f'-> Saving to {out_h5} ...')
        gain_smooth.save_to_h5(out_h5)

        print(f'-> Copy weights ...')
        copy_weights(in_h5, out_h5)

    print('All done !')




#!/usr/bin/env python

import os
import sys

import toml
import click

import numpy as np

import matplotlib as mpl
mpl.use('Agg')

import matplotlib.pyplot as plt

from nenucal import __version__

# Lazy loaded:
# - nenucal.msutils in most commands
# - nenucal.settings, nenucal.flagutils in ssins() and plot_ssins_result()
# - nenucal.delayflag in delay_flag()


mpl.rcParams['image.cmap'] = 'Spectral_r'
mpl.rcParams['image.origin'] = 'lower'
mpl.rcParams['image.interpolation'] = 'nearest'
mpl.rcParams['axes.grid'] = True

t_file = click.Path(exists=True, dir_okay=False)
t_dir = click.Path(exists=True, file_okay=False)


@click.group()
@click.version_option(__version__)
def main():
    ''' Flagging utilities: backup/restore/copy flags; flag with SSINS ...'''


@main.command('backup')
@click.argument('ms_in', type=t_dir)
@click.argument('backup_file')
def backup(ms_in, backup_file):
    ''' Backup FLAG from MS_IN to BACKUP_FILE'''
    from nenucal.msutils import FlagManager

    flag = FlagManager.load_from_ms(ms_in)
    flag.save(backup_file)


@main.command('restore')
@click.argument('ms_in', type=t_dir)
@click.argument('backup_file', type=t_file)
def restore(ms_in, backup_file):
    ''' Restore FLAG from BACKUP_FILE to MS_IN '''
    from nenucal.msutils import FlagManager

    flag = FlagManager.load(backup_file)
    flag.save_to_ms(ms_in)


@main.command('reset')
@click.argument('ms_ins', nargs=-1, type=t_dir)
def reset(ms_ins):
    ''' Reset all FLAG of MS_INS '''
    from nenucal.msutils import FlagManager

    for ms_in in ms_ins:
        FlagManager.reset(ms_in)


@main.command('copy')
@click.argument('ms_in', type=t_dir)
@click.argument('ms_out', type=t_dir)
def copy(ms_in, ms_out):
    ''' Copy FLAG from MS_IN to MS_OUT '''
    from nenucal.msutils import FlagManager

    flag = FlagManager.load_from_ms(ms_in)
    flag.save_to_ms(ms_out)


def plot_ssins_result(ms_filename, ms_data, mask, output_file):
    from nenucal import flagutils

    ss_spc_before = flagutils.get_ss_spc(ms_data.data, zero_mean=False)
    ss_spc_after = flagutils.get_ss_spc(np.ma.array(ms_data.data, mask=mask), zero_mean=False)

    vmin = np.quantile(ss_spc_after, 0.001)
    vmax = np.quantile(ss_spc_after, 0.999)

    fig, (ax1, ax2) = plt.subplots(ncols=2, sharey=True, figsize=(8, 3))
    ax1.imshow(ss_spc_before, aspect='auto', vmin=vmin, vmax=vmax)
    ax2.imshow(ss_spc_after, aspect='auto', vmin=vmin, vmax=vmax)
    ax1.set_ylabel('Freq index')
    ax1.set_xlabel('Time index')
    ax2.set_xlabel('Time index')
    ax1.set_title('Before')
    ax2.set_title('After')
    fig.suptitle(f'SSINS flagging result for {os.path.basename(ms_filename.strip("/"))}', y=1)

    fig.tight_layout()
    fig.savefig(output_file)


def plot_baseline_result(ms_filename, ms_data, mask, output_file):
    from nenucal import msutils

    def get_baseline_mask_map(ant1, ant2, mask):
        idx_freq_full_flag = (mask[:, :, :, 0].sum(axis=(1, 2)) == mask.shape[1] * mask.shape[2])
        ant_max = max([ant2.max(), ant1.max()])
        su = mask[~idx_freq_full_flag].sum(axis=(0, 1, 3))
        si = (mask[~idx_freq_full_flag].shape[0] * mask.shape[1] * mask.shape[3])
        return msutils.make_ant_matrix(ant1, ant2, 100. * su / si, a_max=ant_max)

    fig, (ax1, ax2) = plt.subplots(ncols=2, sharey=True, figsize=(7.5, 3.5))
    ax1.imshow(get_baseline_mask_map(ms_data.ant1, ms_data.ant2, ms_data.data.mask), vmin=0, vmax=100)
    im = ax2.imshow(get_baseline_mask_map(ms_data.ant1, ms_data.ant2, mask), vmin=0, vmax=100)
    plt.colorbar(im, ax=ax2)
    fig.tight_layout()
    ax1.set_ylabel('Ant 1')
    ax1.set_xlabel('Ant 2')
    ax2.set_xlabel('Ant 2')
    ax1.set_title('Before')
    ax2.set_title('After')
    fig.suptitle(f'SSINS flagging result for {os.path.basename(ms_filename.strip("/"))}', y=1)

    fig.tight_layout()
    fig.savefig(output_file)


def backup_restore(ms_in, backup_flag, restore_flag, backup_file):
    from nenucal import msutils

    h5_flag_backup = os.path.join(ms_in, backup_file)

    if backup_flag and restore_flag:
        print('Either backup or restore are allowed (not both).')
        sys.exit(1)

    if backup_flag:
        if os.path.exists(h5_flag_backup):
            print('Original flags have already been backed up')
        else:
            print('Backing up original flags ...')
            msutils.FlagManager.load_from_ms(ms_in).save(h5_flag_backup)

    if restore_flag:
        if not os.path.exists(h5_flag_backup):
            print('Can not restore: backup does not exist.')
            sys.exit(1)
        print('Restoring flags ...')
        msutils.FlagManager.load(h5_flag_backup).save_to_ms(ms_in)
        print(f'Done')
        sys.exit(0)


@main.command('ssins')
@click.argument('ms_in', type=t_dir)
@click.option('--umin', help='Min baseline (in meter)', type=float, default=20, show_default=True)
@click.option('--umax', help='Max baseline (in meter)', type=float, default=4000, show_default=True)
@click.option('--config', help='SSINS configuration file', type=t_file, default=None)
@click.option('--data_col', help='DATA column to run SSINS on', type=str, default='DATA')
@click.option('--dry_run', help='Do not apply flags', is_flag=True)
@click.option('backup_flag', '--backup', help='Backup flags before SSINS', is_flag=True)
@click.option('restore_flag', '--restore', help='Restore previously backup-ed flags', is_flag=True)
@click.option('--plot_dir', help='Plot directory', type=t_dir, default=None)
@click.option('--backup_file', help='Backup filename', type=str, default='flags_before_ssins.h5')
def ssins(ms_in, umin, umax, data_col, config, dry_run, backup_flag, restore_flag, plot_dir, backup_file):
    ''' Run the SSINS flagger algorithm on MS_IN '''
    from nenucal import msutils
    from nenucal import flagutils, settings

    if config is None:
        config = os.path.join(settings.TEMPLATE_DIR, 'ssins_config.toml')

    config = toml.load(config)

    backup_restore(ms_in, backup_flag, restore_flag, backup_file)

    print(f'Opening MS file {ms_in} ...')
    ms = msutils.MsDataCube.load(ms_in, umin, umax, data_col)

    if ms.weights is not None:
        ms.data = ms.data * (ms.weights) ** .5

    print(f'Flag before SSINS: {100. * ms.data.mask.sum() / ms.data.mask.size:.2f} %')

    print(f'Starting SSINS flagging ...')
    mask_ssins = np.sum([flagutils.ssins_flagger(ms.data, k, config) for k in config["n_time_avg"]], axis=0)
    mask = (mask_ssins + ms.data.mask).astype(bool)

    mask_time_freq = flagutils.time_freq_threshold_flagger(mask, config["time_freq_threshold"])
    mask_baseline = flagutils.baseline_threshold_flagger(mask, config["baseline_threshold"])
    mask_snaphot = flagutils.snapshot_threshold_flagger((mask + mask_time_freq + mask_baseline).astype(bool),
                                                        config["snapshot_threshold"])

    mask = (mask_ssins + ms.data.mask + mask_time_freq + mask_baseline + mask_snaphot).astype(bool)

    if plot_dir is not None:
        print(f'Plotting results ...')

        if not os.path.isdir(plot_dir):
            os.makedirs(plot_dir)

        file = os.path.basename(ms_in.strip('/')).split('.MS')[0]
        plot_ssins_result(ms_in, ms.data, mask, os.path.join(plot_dir, f'ssins_flag_{file}.png'))
        plot_baseline_result(ms_in, ms, mask, os.path.join(plot_dir, f'baseline_flag_{file}.png'))

    print(f'Flag after SSINS: {100. * mask.sum() / mask.size:.2f} %')

    if not dry_run:
        print(f'Saving to MS file {ms_in} ...')

        ms.save_flag(ms_in, mask)

    print(f'All done')


@main.command('delay_flagger')
@click.argument('ms_in', type=t_dir)
@click.option('--data_col', help='DATA column to run delay flagger on', type=str, default='DATA')
@click.option('--umin', help='Min baseline (in meter)', type=float, default=50, show_default=True)
@click.option('--umax', help='Max baseline (in meter)', type=float, default=400, show_default=True)
@click.option('--n_time_avg', help='Number of time slot to averages', type=int, default=50)
@click.option('--n_times', help='Number of time bins', type=int, default=20)
@click.option('--n_sigma_i', help='Stokes I n sigma threshold', type=int, default=6)
@click.option('--n_sigma_v', help='Stokes V n sigma threshold', type=int, default=6)
@click.option('backup_flag', '--backup', help='Backup flags before delay flagger', is_flag=True)
@click.option('restore_flag', '--restore', help='Restore previously backup-ed flags', is_flag=True)
@click.option('--plot_dir', help='Plot directory', type=t_dir, default=None)
@click.option('--plot_ps_all_baselines', is_flag=True)
@click.option('--backup_file', help='Backup filename', type=str, default='flags_before_delay_flag.h5')
@click.option('--dry_run', help='Do not apply flags', is_flag=True)
def delay_flagger(ms_in, data_col, umin, umax, n_time_avg, n_times, n_sigma_i, n_sigma_v,
                  backup_flag, restore_flag, plot_dir, plot_ps_all_baselines, backup_file, dry_run):
    ''' Run the delay flagger algorithm on MS_IN '''
    from nenucal import delayflag

    backup_restore(ms_in, backup_flag, restore_flag, backup_file)

    f = delayflag.DelayFlagger(ms_in, n_time_avg, data_col, umin, umax)
    res = f.do_flag(n_times, n_sigma_i, n_sigma_v)

    if plot_dir is not None:
        print(f'Plotting results ...')
        if plot_ps_all_baselines:
            f.plot_delay_ps_iv(plot_dir)
        f.plot_delay_flag(res, plot_dir)

    if not dry_run:
        f.save_flag(res)


if __name__ == '__main__':
    main()



import os

import numpy as np

import scipy.signal
import scipy.ndimage
import scipy.stats

import astropy.constants as const
import astropy.stats as astats
import astropy.time as at

from . import utils, msutils

import matplotlib as mpl
mpl.use('Agg')

import matplotlib.pyplot as plt


def flag_time(ms_cube, threshold=0.15):
    s = ms_cube.data.shape
    mask_freqs = (ms_cube.data.mask.sum(axis=(0, 3)) > threshold * (s[0] * s[3]))[None, :, :, None]
    ms_cube.data.mask = (ms_cube.data.mask + mask_freqs).astype(bool)
    ms_cube.data_dt.mask = (ms_cube.data_dt.mask + mask_freqs[:, ::2][:, :ms_cube.data_dt.mask.shape[1]]).astype(bool)


def hpass_flag(data, fill=0):
    d = data.filled(fill)
    k = scipy.signal.gaussian(150, 20)
    fit_r = scipy.ndimage.filters.convolve1d(d.real, k / k.sum(), mode='nearest', axis=0)
    fit_i = scipy.ndimage.filters.convolve1d(d.imag, k / k.sum(), mode='nearest', axis=0)
    fit = (fit_r + 1j * fit_i)
    return np.ma.array(d - fit, mask=data.mask)


def get_lst(obs_mjd, longitude=6.57):
    return at.Time(obs_mjd, scale='utc', format='mjd').sidereal_time('mean', longitude=longitude).value


def binned_sigma_clip(ru, d, bins=[0, 250], detrend_deg=0, **kargs):
    y = d.flatten()
    if detrend_deg > 0:
        detrend_y = astats.sigma_clip(y, sigma=6)
        d_fit = np.poly1d(np.polyfit(ru[~detrend_y.mask], y[~detrend_y.mask], detrend_deg))(ru)
        y = y - d_fit

    r = np.ma.zeros(y.shape)
    indices = np.digitize(ru, bins)
    for i in np.unique(indices):
        r[indices == i] = astats.sigma_clip(y[indices == i], **kargs)
    return np.ma.array(d, mask=r.mask.reshape(*d.shape))


def make_ps(d, freq, half=True, window_fct='blackmanharris'):
    s = d.shape
    w = scipy.signal.get_window(window_fct, s[0])[:, None]
    w = w / w.mean()

    if len(s) > 2:
        d = d.reshape((s[0], s[1] * s[2]))

    _, tf = utils.nudft(freq, utils.rmean(d), s[0], w=w)

    if len(s) > 2:
        tf = tf.reshape((s[0], s[1], s[2]))

    ps = abs(tf) ** 2

    if half:
        M = ps.shape[0]
        if utils.is_odd(M):
            ps = 0.5 * (ps[M // 2 + 1:] + ps[:M // 2][::-1])
        else:
            ps = 0.5 * (ps[M // 2 + 1:] + ps[1:M // 2][::-1])

    return np.ma.array(ps, mask=ps == 0)


class DelayFlaggerResult(object):

    def __init__(self, n_times, n_sigma_i, n_sigma_v, all_m_i_masked, all_m_v_masked):
        self.n_times = n_times
        self.n_sigma_i = n_sigma_i
        self.n_sigma_v = n_sigma_v
        self.all_m_i_masked = all_m_i_masked
        self.all_m_v_masked = all_m_v_masked


class DelayFlagger(object):

    def __init__(self, ms_file, ntime_avg, data_col='DATA', umin=50, umax=400):
        self.ms_file = ms_file
        self.ms_cube = msutils.MsDataCube.load(ms_file, umin, umax, data_col, n_time_avg=1)

        flag_time(self.ms_cube)
        m_vis_i, m_vis_v, m_vis_dt = self._get_stokes(ntime_avg)

        half = False
        window_fct = 'blackmanharris'

        self.all_ps_dt = make_ps(m_vis_dt.filled(0), self.ms_cube.freq, half=half, window_fct=window_fct)
        self.all_ps_v = make_ps(m_vis_v.filled(0), self.ms_cube.freq, half=half, window_fct=window_fct)
        self.all_ps_i = make_ps(hpass_flag(m_vis_i).filled(0), self.ms_cube.freq, half=half, window_fct=window_fct)

        self.delay = utils.get_delay(self.ms_cube.freq * 1e-6, half=half)

        freq = self.ms_cube.freq.mean()
        self.lamb = const.c.value / freq
        self.horizon = utils.get_wedge_delay(np.radians(90), self.ms_cube.bu / self.lamb, freq) * 1e6

    def _get_stokes(self, ntime_avg=50):
        m_vis_dt = self.ms_cube.data_dt[:, :, :, 0]
        m_vis_v = 0.5 * (-1j * (self.ms_cube.data[:, :, :, 1] - self.ms_cube.data[:, :, :, 2]))
        m_vis_i = 0.5 * (self.ms_cube.data[:, :, :, 0] + self.ms_cube.data[:, :, :, 3])

        m_vis_i, m_vis_i_n = msutils.mean_consecutive(m_vis_i, axis=1, n=ntime_avg, return_n=True)
        m_vis_i = m_vis_i * np.sqrt(m_vis_i_n)

        m_vis_v, m_vis_v_n = msutils.mean_consecutive(m_vis_v, axis=1, n=ntime_avg, return_n=True)
        m_vis_v = m_vis_v * np.sqrt(m_vis_v_n)

        m_vis_dt, m_vis_dt_n = msutils.mean_consecutive(m_vis_dt, axis=1, n=ntime_avg, return_n=True)
        m_vis_dt = m_vis_dt * np.sqrt(m_vis_dt_n)

        return m_vis_i, m_vis_v, m_vis_dt

    def do_flag(self, n_times, n_sigma_i, n_sigma_v):
        all_m_i_masked = self.filter_delay_ps(self.all_ps_i, n_times, n_sigma_i)
        all_m_v_masked = self.filter_delay_ps(self.all_ps_v, n_times, n_sigma_v)

        all_m_i_masked_1 = self.filter_delay_ps(self.all_ps_i, 1, n_sigma_i)
        all_m_v_masked_1 = self.filter_delay_ps(self.all_ps_v, 1, n_sigma_v)

        all_m_i_masked.mask = all_m_i_masked.mask + all_m_i_masked_1.mask
        all_m_v_masked.mask = all_m_v_masked.mask + all_m_v_masked_1.mask

        return DelayFlaggerResult(n_times, n_sigma_i, n_sigma_v, all_m_i_masked, all_m_v_masked)

    def save_flag(self, result):
        new_mask = self.ms_cube.data.mask.copy()
        print('Before:', new_mask.sum() / self.ms_cube.data.mask.size)

        all_m_mask = result.all_m_i_masked.mask + result.all_m_v_masked.mask
        idx = list(np.linspace(0, all_m_mask.shape[1] - 1, len(self.ms_cube.time)).astype(int))
        new_mask = (new_mask + all_m_mask[:, idx].T[None, :, :, None]).astype(bool)
        print('After flagging:', new_mask.sum() / self.ms_cube.data.mask.size)

        self.ms_cube.save_flag(self.ms_file, new_mask)

    def get_delay_power_ratio_map(self, all_ps, n_times):
        all_m = []
        for i in np.arange(len(self.ms_cube.ru)):
            mean = np.mean(self.all_ps_dt[:, :, i] / 4.)
            m = np.mean(all_ps[(abs(self.delay) > 1.2 * self.horizon[i]) & (abs(self.delay) < 4), :, i], axis=0) / mean
            if np.alltrue(m.mask):
                all_m.append(np.ones((n_times)) * np.nan)
                continue
            m[m.mask] = np.nanmean(m)
            if n_times is None:
                all_m.append(m)
            else:
                all_m.append(scipy.stats.binned_statistic(np.arange(len(m)), m.filled(0),
                                                          statistic=np.ma.mean, bins=n_times)[0])

        return np.array(all_m)

    def filter_delay_ps(self, all_ps, n_times, n_sigma):
        all_m = self.get_delay_power_ratio_map(all_ps, n_times)
        ru = np.repeat(self.ms_cube.ru, n_times)

        return binned_sigma_clip(ru, all_m, sigma=n_sigma, stdfunc=astats.mad_std)

    def plot_delay_flag(self, result, plot_dir):
        cmap = mpl.cm.get_cmap('magma')
        cmap.set_bad(color='blue')
        props = dict(boxstyle='round', facecolor='white', alpha=0.8)
        extent = [0, len(self.ms_cube.ru) - 1, 0, len(self.ms_cube.time) - 1]
        lsts = get_lst(self.ms_cube.time[:] / 3600. / 24.)

        fig, (axi1, axi2, axv1, axv2) = plt.subplots(nrows=4, figsize=(12, 11), sharex=True)

        for ax1, ax2, all_m_masked, stokes in [(axi1, axi2, result.all_m_i_masked, 'I'),
                                               (axv1, axv2, result.all_m_v_masked, 'V')]:
            im = ax1.imshow(all_m_masked.data[np.argsort(self.ms_cube.ru)].T,
                            aspect='auto', vmax=2, vmin=0.9, cmap=cmap, extent=extent)
            plt.colorbar(im, ax=ax1)
            ax1.set_ylabel('LST Time (hour)')
            ax1.set_yticks(np.linspace(0, len(self.ms_cube.time) - 1, result.n_times))
            ax1.set_yticklabels(['%.1f' % lsts[int(i)] for i in ax1.get_yticks()])

            txt = f'Stokes {stokes}: Max: {np.nanmax(all_m_masked.data):.1f}, \
            Med: {np.nanmedian(all_m_masked.data):.2f}'

            ax1.text(0.05, 0.95, txt, transform=ax1.transAxes, va='top', ha='left', bbox=props)

            ax2.imshow(all_m_masked[np.argsort(self.ms_cube.ru)].T, aspect='auto',
                       vmax=2, vmin=0.9, cmap=cmap, extent=extent)
            plt.colorbar(im, ax=ax2)
            ax2.set_ylabel('LST Time (hour)')
            ax2.set_xlabel('Baseline number')
            ax2.set_yticks(np.linspace(0, len(self.ms_cube.time) - 1, result.n_times))
            ax2.set_yticklabels(['%.1f' % lsts[int(i)] for i in ax2.get_yticks()])

            txt = f'Stokes {stokes}: Max: {all_m_masked.max():.1f}, Med: \
            {np.ma.median(all_m_masked):.2f}, Flagged: {all_m_masked.mask.sum() / all_m_masked.size * 100:.1f} %'

            ax2.text(0.05, 0.95, txt, transform=ax2.transAxes, va='top', ha='left', bbox=props)

            for ax in [ax1, ax2]:
                for l in [50, 100, 150]:
                    x = np.where(self.ms_cube.ru[np.argsort(self.ms_cube.ru)] / self.lamb > l)[0]
                    if len(x) > 0:
                        ax.axvline(x[0], c='red', ls='--')
                        ax.text(x[0], 0, str(l), c='red')

        fig.tight_layout()

        if not os.path.exists(plot_dir):
            os.makedirs(plot_dir)

        fig.savefig(plot_dir + f'/all_delay_ps_mask_i{result.n_sigma_i}_v{result.n_sigma_v}.pdf')

    def plot_delay_ps_iv(self, plot_dir):
        # stat_names = self.ms_cube.ant_name
        stat_names = np.array(['STA' + str(k) for k in np.arange(56)])
        nrows = int(np.ceil(self.all_ps_i.shape[2] / 5.))
        fig, axs = plt.subplots(ncols=5, nrows=nrows, sharex=True, sharey=True, figsize=(12, 1 + nrows * 2))

        all_noise_levels = []
        props = dict(boxstyle='round', facecolor='white', alpha=0.6)

        for i_ax, i in enumerate(np.argsort(self.ms_cube.ru)):
            ax = axs.flatten()[i_ax]
            txt = '%.2f / %s - %s' % (self.ms_cube.ru[i] / self.lamb, stat_names[self.ms_cube.ant1[i]]
                                      [3:], stat_names[self.ms_cube.ant2[i]][3:])

            if np.alltrue(self.ms_cube.data[:, :, i, 0].mask):
                all_noise_levels.append((np.nan, np.nan))
                ax.text(0.05, 0.95, txt, transform=ax.transAxes, va='top', ha='left', bbox=props)
                continue

            mean = np.mean(self.all_ps_dt[:, :, i] / 4.)

            ax.plot(self.delay, np.ma.mean(self.all_ps_v[:, :, i], axis=1))
            ax.plot(self.delay, np.ma.mean(self.all_ps_i[:, :, i], axis=1))
            ax.axvline(self.horizon[i], ls='--', c=utils.black)
            ax.axvline(- self.horizon[i], ls='--', c=utils.black)
            ax.axhline(mean, ls='--', c=utils.black)
            ax.set_yscale('log')
            ax.set_xlim(-4, 4)

            noise_r_i = np.mean(self.all_ps_i[(abs(self.delay) > 1.2 * self.horizon[i]) & (abs(self.delay) < 4), :, i]) / mean
            noise_r_v = np.mean(self.all_ps_v[(abs(self.delay) > 1.2 * self.horizon[i]) & (abs(self.delay) < 4), :, i]) / mean
            all_noise_levels.append((noise_r_i, noise_r_v))

            color = 'black'

            txt += '\n%s: I:%.1f V:%.1f' % (i, noise_r_i, noise_r_v)
            ax.text(0.05, 0.95, txt, transform=ax.transAxes, va='top', ha='left', bbox=props, color=color)

        fig.tight_layout(pad=0)
        fig.savefig(plot_dir + '/all_delay_ps_iv.pdf')



#!/usr/bin/env python

import click

import numpy as np

import astropy.units as u
from astropy.time import Time
from astropy.coordinates import SkyCoord

from casacore import tables

from nenucal import __version__
from nenucal import skymodel


t_file = click.Path(exists=True, dir_okay=False)
t_dir = click.Path(exists=True, file_okay=False)


@click.group()
@click.version_option(__version__)
def main():
    ''' NenuFAR sky model utilities: Build sky model from catalog; transform intrinsic to apparent sky model '''


@main.command('build')
@click.argument('ms_file', type=t_dir)
@click.option('--catalog', '-c', help='The catalog to fetch data from',
              type=click.Choice(['specfind', 'lcs165']), default='lcs165')
@click.option('--min_flux', '-m', help='Min Flux of the catalog at 74 MHz in Jy', default=0.5,
              type=float, show_default=True)
@click.option('--radius', '-r', help='Radius around the phase center', default=20, type=float, show_default=True)
@click.option('--out_file', '-o', help='Output filename', default='catalog.skymodel', type=str, show_default=True)
def build(ms_file, catalog, min_flux, radius, out_file):
    ''' Build an intrinsic sky model from catalog
    '''
    skymodel.build_sky_model_ms(ms_file, min_flux, radius, out_file, catalog=catalog)


@main.command('attenuate')
@click.argument('ms_file', type=t_dir)
@click.argument('input_models', type=t_file, nargs=-1, required=True)
@click.option('--min_flux', '-m', help='Minimum flux of the apparent sky model', type=float,
              default=0.5, show_default=True)
@click.option('--min_flux_patch', '-p', help='Minimum patch flux', type=float, default=20, show_default=True)
@click.option('--min_elevation', '-e', help='Minimum elevation in degree', type=float, default=10, show_default=True)
@click.option('--keep', '-k', multiple=True, help='Always keep specified patch')
@click.option('--remove', '-r', multiple=True, help='Always remove specified patch')
@click.option('--out_file', '-o', help='Output filename', default='apparent.skymodel', type=str, show_default=True)
def attenuate(ms_file, input_models, min_flux, min_flux_patch, min_elevation, keep, remove, out_file):
    ''' Make apparent sky model from intrinsic beam model
    '''
    ms = tables.table(ms_file)
    time = ms.getcol('TIME')
    mtime = np.array([np.quantile(time, 0.25), np.quantile(time, 0.5), np.quantile(time, 0.75)])
    mfmhz = tables.table(ms.getkeyword('SPECTRAL_WINDOW')).getcol('CHAN_FREQ').mean() * 1e-6
    phase_dir = tables.table(ms.getkeyword('FIELD')).getcol('PHASE_DIR').squeeze()

    coord_phase_dir = SkyCoord(ra=phase_dir[0] * u.rad, dec=phase_dir[1] * u.rad)
    observing_time = Time(mtime / 3600. / 24., scale='utc', format='mjd')

    sky_model = skymodel.concatenate(input_models)

    sky_model = skymodel.apply_nenufar_beam(sky_model, observing_time, coord_phase_dir, mfmhz)

    sky_model = skymodel.edit_model(sky_model, observing_time, min_flux=min_flux, min_flux_patch=min_flux_patch, 
                                    min_elevation_patch=min_elevation, always_keep=keep, always_remove=remove)

    print(f'Saving apparent sky model to {out_file} ...')
    sky_model.write(out_file, clobber=True)
    print('All done !')


@main.command('concatenate')
@click.argument('ms_file', type=t_dir)
@click.argument('input_models', type=t_file, nargs=-1, required=True)
@click.option('--min_flux', '-m', help='Minimum flux of the apparent sky model', type=float,
              default=0.5, show_default=True)
@click.option('--min_flux_patch', '-p', help='Minimum patch flux', type=float, default=20, show_default=True)
@click.option('--min_elevation', '-e', help='Minimum elevation in degree', type=float, default=10, show_default=True)
@click.option('--keep', '-k', multiple=True, help='Always keep specified patch')
@click.option('--remove', '-r', multiple=True, help='Always remove specified patch')
@click.option('--out_file', '-o', help='Output filename', default='apparent.skymodel', type=str, show_default=True)
def concatenate(ms_file, input_models, min_flux, min_flux_patch, min_elevation, keep, remove, out_file):
    ''' Concatenate multiple sky models file into one
    '''
    ms = tables.table(ms_file)
    time = ms.getcol('TIME')
    mtime = np.array([np.quantile(time, 0.25), np.quantile(time, 0.5), np.quantile(time, 0.75)])
    observing_time = Time(mtime / 3600. / 24., scale='utc', format='mjd')

    sky_model = skymodel.concatenate(input_models)

    sky_model = skymodel.edit_model(sky_model, observing_time, min_flux=min_flux, min_flux_patch=min_flux_patch, 
                                    min_elevation_patch=min_elevation, always_keep=keep, always_remove=remove)

    print(f'Saving concatenated sky model to {out_file} ...')
    sky_model.write(out_file, clobber=True)
    print('All done !')


@main.command('set_patch_name')
@click.argument('input_model', type=t_file)
@click.argument('patch_name', type=str)
@click.option('--out_file', '-o', help='Output filename', default=None, type=str, show_default=True)
def set_patch_name(input_model, patch_name, out_file):
    ''' Set a single patch name in the sky model
    '''

    if not out_file:
        out_file = input_model

    sky_model = skymodel.lsmtool.load(input_model)
    skymodel.set_patch_name(sky_model, patch_name)
    sky_model.write(out_file, clobber=True)

if __name__ == '__main__':
    main()



#!/usr/bin/env python

import os
import sys
import fnmatch
from collections import defaultdict

import toml
import click

from nenucal import tasks, datahandler, msutils, __version__
from nenucal.settings import Settings, ImgSettings

from libpipe import worker


t_file = click.Path(exists=True, dir_okay=False)
t_dir = click.Path(exists=True, file_okay=False)

WSCLEAN_BIN = 'wsclean'


def build_wsclean_cmd(s, ms_in):
    cmd = [WSCLEAN_BIN]
    img_dir = f'{ms_in}/images'
    for key, value in s.items():
        if key == 'name':
            value = f'{img_dir}/{value}'

        if key == 'channels-out':
            n_channels = len(msutils.get_ms_freqs(ms_in)[0])
            if value == 'all':
                value = n_channels
            elif str(value).startswith('every'):
                value = n_channels // int(value[5:])

        if not os.path.exists(img_dir):
            os.makedirs(img_dir)

        if isinstance(value, bool):
            cmd.append(f'-{key}')
        else:
            cmd.append(f'-{key} {value}')

    return ' '.join(cmd)


@click.command()
@click.version_option(__version__)
@click.argument('config_file', type=click.Path(exists=True, dir_okay=False))
@click.argument('ms_ins_or_obs_ids', nargs=-1, required=True)
@click.option('--combine', '-c', help='Combine all MS to produce one image (for each SW when applicable).', is_flag=True)
def main(config_file, ms_ins_or_obs_ids, combine):
    ''' Calibration pipeline

        \b
        CONFIG_FILE: Configuration file
        MS_INS_OR_OBS_IDS: Measurement sets to process or OBS_IDS in case you have set data_handler in CONFIG_FILE
    '''
    try:
        s = ImgSettings.load_with_defaults(config_file, check_args=False)
    except toml.TomlDecodeError as e:
        print(f'Error parsing configuration: {e}')
        sys.exit(1)
    
    ms_ins = defaultdict(list)

    if 'data_handler' in s and "config_file" in s["data_handler"] and s["data_handler"]["config_file"]:
        print(f'Using {s["data_handler"]["config_file"]} data handler configuration')
        dh = datahandler.DataHandler.from_file(s['data_handler']['config_file'])
        level = s['data_handler']['data_level']
        for obs_id_str in ms_ins_or_obs_ids:
            obs_ids, sws = dh.get_obs_ids_and_spectral_windows(obs_id_str)
            for sw in sws:
                for obs_id in obs_ids:
                    ms_ins[sw].extend(dh.get_ms_path(obs_id, level, sw))
    else:
        ms_ins['SW0'] = ms_ins_or_obs_ids

    w = worker.get_worker_pool('Imaging', nodes=s.worker.nodes, env_file=s.worker.env_file, 
                               max_concurrent=s.worker.max_concurrent, debug=s.worker.debug, 
                               dry_run=s.worker.dry_run)

    for sw in ms_ins.keys():
        if combine:
            base_cmd = build_wsclean_cmd(s.wsclean, ms_ins[sw][-1])
            host = s.get_target_host(ms_ins[sw][-1])
            cmd = f'{base_cmd} {" ".join(ms_ins[sw])}'
            w.add(cmd, run_on_host=host)
        else:
            for ms in ms_ins[sw]:
                base_cmd = build_wsclean_cmd(s.wsclean, ms)
                host = s.get_target_host(ms)
                cmd = f'{base_cmd} {ms}'
                w.add(cmd, run_on_host=host)

    w.execute()

    print('All done !')


if __name__ == '__main__':
    main()




# Mixed utilities
#
# Author: F. Mertens

import numpy as np


color_cycle = ['#3465a4', '#4e9a06', '#cc0000', '#F57900', '#75507b', '#EDD400', '#555753']
blue, green, red, orange, magenta, yellow, black = color_cycle


def chunks(arr, n_max):
    n = int(np.ceil(len(arr) / np.ceil(len(arr) / n_max)))
    for i in range(0, len(arr), n):
        yield arr[i:i + n]


def factors(n):
    r = np.arange(1, int(n ** 0.5) + 1)
    x = r[np.mod(n, r) == 0]
    return np.unique(np.concatenate((x, n / x), axis=None))


def rmean(data, axis=0):
    '''Remove the mean in direction axis'''
    return data - np.mean(data, axis=axis)


def is_odd(num):
    return num & 0x1


def robust_freq_width(freqs):
    '''Return frequency width, robust to gaps in freqs'''
    dfs = np.diff(freqs)
    m, idx, c = np.unique(np.round(dfs * 1e-3) * 1e3, return_counts=True, return_inverse=True)
    return dfs[np.where(idx == np.argmax(c))].mean()


def nudft(x, y, M=None, w=None, dx=None):
    """Non uniform discrete Fourier transform

    Args:
        x (array): x axis
        y (array): y axis
        M (int, optional): Number of Fourier components that will be computed, default to len(x)
        w (array, optional): Tapper

    Returns:
        (array, array): k modes, Fourier transform of y
    """
    if M is None:
        M = len(x)

    if dx is None:
        dx = robust_freq_width(x)

    if w is not None:
        y = y * w

    df = 1 / (dx * M)
    k = df * np.arange(-(M // 2), M - (M // 2))

    X = np.exp(2 * np.pi * 1j * k * x[:, np.newaxis])

    return k, np.tensordot(y, X.conj().T, axes=[0, 1]).conj().T


def get_delay(freqs, M=None, dx=None, half=True):
    ''' Convert frequencies to delay '''
    if dx is None:
        dx = robust_freq_width(freqs)
    if M is None:
        M = len(freqs)

    df = 1 / (dx * M)
    delay = df * np.arange(-(M // 2), M - (M // 2))

    if half:
        M = len(delay)
        delay = delay[M // 2 + 1:]

    return delay


def get_wedge_delay(theta, bu, freq):
    return np.sin(theta) * bu / freq



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['nenucal', 'nenucal.tests', 'nenucal.tools']

package_data = \
{'': ['*'], 'nenucal': ['cal_config/*', 'templates/*']}

install_requires = \
['GPy>=1.9.9,<2.0.0',
 'astropy>=4.0,<6.0',
 'astroquery>=0.4',
 'click>=7.0,<9.0',
 'libpipe>=0.1.1',
 'losoto>=2.3,<3.0',
 'lsmtool>=1.4.2,<2.0.0',
 'matplotlib>=3.0,<4.0',
 'nenupy>=1.0,<2.0',
 'python-casacore>=3.0,<4.0',
 'requests>=2.0,<3.0',
 'scipy>=1.4,<2.0',
 'tables>=3.2,<4.0',
 'tabulate>=0.8',
 'toml>=0.10']

extras_require = \
{':python_version < "3.8"': ['importlib-metadata>=1.0,<2.0']}

entry_points = \
{'console_scripts': ['calpipe = nenucal.tools.calpipe:main',
                     'flagtool = nenucal.tools.flagtool:main',
                     'imgpipe = nenucal.tools.imgpipe:main',
                     'modeltool = nenucal.tools.modeltool:main',
                     'nenudata = nenucal.tools.nenudata:main',
                     'soltool = nenucal.tools.soltool:main']}

setup_kwargs = {
    'name': 'nenucal',
    'version': '0.2',
    'description': 'Calibration pipeline for the NenuFAR Cosmic Dawn project',
    'long_description': 'NenuCAL CD\n==========\n\nCalibration pipeline and set of utilities for the NenuFAR Cosmic Dawn project. The calibration uses the standard LOFAR [DPPP](https://github.com/lofar-astron/DP3/tree/master/DPPP) in the background and should be general enough to be used by any NenuFAR projects.\n\nInstallation\n------------\n\nnenucal-cd can be installed via pip:\n\n    $ pip install nenucal-cd\n\nand requires Python 3.7.0 or higher. For most tasks, you will also need [DPPP](https://github.com/lofar-astron/DP3/tree/master/DPPP), [LSMTool](https://www.astron.nl/citt/lsmtool/overview.html) that you can install with the command:\n\n    $ pip install git+https://github.com/darafferty/LSMTool.git\n\nand [LoSoTo](https://revoltek.github.io/losoto/) that you can install with the command:\n\n    $ pip install git+https://github.com/revoltek/losoto\n\nUsage\n-----\n\nCalibration is performed using the `calpipe` utility. You will find [here some documentations](https://gitlab.com/flomertens/nenucal-cd/-/wikis/Calibrating-NenuFAR-data) on how to use it.\n\nAdditionally to calpipe, this package provides a tool to manage the FLAG column and run a SSINS flagger ([flagtool](https://gitlab.com/flomertens/nenucal-cd/-/wikis/flagging-utility)), a tool to plot and smooth gain solutions ([soltool](https://gitlab.com/flomertens/nenucal-cd/-/wikis/soltool-utility)) and a tool to produce apparent sky model from catalog ([modeltool](https://gitlab.com/flomertens/nenucal-cd/-/wikis/sky-model-utility)).\n\nContact\n-------\n\nDo not hesitate to fill an [issue](https://gitlab.com/flomertens/nenucal-cd/-/issues) or contact me if you encounter a problem with this package.\n',
    'author': '"Florent Mertens"',
    'author_email': '"florent.mertens@gmail.com"',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.com/flomertens/nenucal-cd',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import numpy as np

from nenucal import utils


def test_chunks():
    for i in range(50):
        l = np.arange(np.random.randint(1, 200))
        n_max = np.random.randint(1, 100)
        chunks = list(utils.chunks(l, n_max))
        print(len(l), n_max, len(chunks))
        assert len(chunks) >= np.floor(len(l) / n_max)
        assert len(chunks) <= np.ceil(len(l) / n_max)
        assert np.allclose(np.concatenate(chunks), l)
